# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Utilities for the compatibility of ODL with the python array API"""

from types import ModuleType
from dataclasses import dataclass
from typing import Callable, Union

__all__ = (
    'ArrayBackend', 
    'lookup_array_backend',
    'get_array_and_backend',
    'check_device',
    'can_cast',
    )


_registered_array_backends = {}

# The backends shipped with ODL, with the dependencies needed to enable them.
standard_known_backends = { 'numpy': ['numpy']
                          , 'pytorch': ['torch'] }

@dataclass
class ArrayOperation:
    name: str
    operation_call: Callable
    supports_single_input: bool
    supports_two_inputs: bool
    supports_out_argument: bool

@dataclass
class ArrayBackend:
    """
    Class to implement the array backend associated to each TensorSpace Implementations.
    
    Attributes
    ----------
    impl : str
        The implementation of the backend, e.g 'numpy'
    array_namespace : ModuleType
        The actual namespace of the backend, e.g np
    available_dtypes : dict
        A dictionnary mapping a Number/str datatype to the corresponding backend-specific datatype, e.g {float:np.float64, 'float64', np.float64, ...}
    array_type : type
        The type of the array once implemented by the backend, e.g np.ndarray
    array_constructor : Callable
        The function the backend uses to create an array, e.g np.asarray
    from_dlpack : Callable
        Stand-in for the `from_dlpack` method of the Python Array API. We would rather use that directly,
        but there are multiple inconsistencies of its behaviour particular in different PyTorch versions,
        so we need to wrap it as a workaround.
    make_contiguous : Callable
        The function the backend uses to make an array contiguous, e.g np.ascontiguousasarray
    identifier_of_dtype : Callable
        The function used to get a string representation of a backend-specific dtype
    available_devices : list[str]
        List of devices accepted by the backend 
    to_cpu : Callable
        Function to copy an array to the CPU
    to_numpy: Callable
        Function to create a Numpy version of an array

    """
    impl: str
    array_namespace: ModuleType
    available_dtypes: dict[str, object]
    array_type: type
    array_constructor: Callable
    from_dlpack: Callable
    make_contiguous: Callable
    identifier_of_dtype: Callable
    available_devices : list[str]
    to_cpu : Callable
    to_numpy: Callable
    to_device: Callable
    def __post_init__(self):
        if self.impl in _registered_array_backends:
            raise KeyError(f"An array-backend with the identifier {self.impl} is already registered. Every backend needs to have a unique identifier.")
        _registered_array_backends[self.impl] = self
        self._array_operations = {}
    def get_dtype_identifier(self, **kwargs) -> str:
        """
        Method for getting a dtype_identifier (str) from an array or a dtype. 
        This is used to retrieve the dtype of a custom object as a string and pass it to another backend.
        The dtype must actually be a dtype object pertaining to the `self` backend.
        Strings or Python types are not allowed here.
        Use `odl.core.util.dtype_utils._universal_dtype_identifier` for a general conversion from
        dtype-ish objects to identifiers.

        Parameters
        ----------
        **kwargs : 'array' or 'dtype'
            This function inputs either an array OR a dtype
        
        Returns
        -------
        dtype_identifier (str)

        Examples
        --------
        >>> backend = odl.lookup_array_backend('numpy')
        >>> backend.get_dtype_identifier(array=np.zeros(10))
        'float64'
        >>> backend.get_dtype_identifier(array=np.zeros(10, dtype = 'float32'))
        'float32'
        >>> backend.get_dtype_identifier(array=np.zeros(10, float))
        'float64'
        >>> backend.get_dtype_identifier(dtype=np.dtype('float64'))
        'float64'
        >>> backend.get_dtype_identifier(dtype=np.zeros(10, dtype = 'float32').dtype)
        'float32'
        >>> backend.get_dtype_identifier(dtype=np.dtype(float))
        'float64'
        >>> backend.get_dtype_identifier(dtype=np.dtype(float), array=np.zeros(10, float))
        Traceback (most recent call last):
        AssertionError: "array and dtype are mutually exclusive parameters"
        >>> backend.get_dtype_identifier(np.dtype(float))
        Traceback (most recent call last):
        TypeError: "ArrayBackend.get_dtype_identifier() takes 1 positional argument but 2 were given"
        """
        if 'array' in kwargs:
            assert 'dtype' not in kwargs, "array and dtype are mutually exclusive parameters"
            return self.identifier_of_dtype(kwargs['array'].dtype)
        if 'dtype' in kwargs:
            assert 'array' not in kwargs, "array and dtype are mutually exclusive parameters"
            return self.identifier_of_dtype(kwargs['dtype'])
        raise ValueError("Either 'array' or 'dtype' argument must be provided.")

    def _probe_elementwise_operation(self, operation):
        """
        Attempt to use a low-level operation in this backend. If successful, the operation is
        then registered in the `_array_operations` dict in a suitable manner."""
        fn = getattr(self.array_namespace, operation)
        test_inputs = { dtk: self.array_constructor([1,2,3], dtype=dtype)
                         for dtk, dtype in self.available_dtypes.items() }
        test_output = None
        supports_single_input = supports_two_inputs = supports_out_argument = False
        for dtype, test_input in test_inputs.items():
            try:
                test_output = fn(test_input)
                supports_single_input = True
            except (TypeError, RuntimeError):
                pass
            try:
                test_output = fn(test_input, test_input)
                supports_two_inputs = True
            except (TypeError, RuntimeError):
                pass
            try:
                if supports_single_input:
                    fn(test_input, out=test_output)
                    supports_out_argument = True
                elif supports_two_inputs:
                    fn(test_input, test_input, out=test_output)
                    supports_out_argument = True
            except (TypeError, RuntimeError):
                pass
            if supports_single_input or supports_two_inputs:
                self._array_operations[operation] = ArrayOperation(
                         name = operation,
                         operation_call = fn,
                         supports_single_input = supports_single_input,
                         supports_two_inputs = supports_two_inputs,
                         supports_out_argument = supports_out_argument)
                return

    def lookup_array_operation(self, operation: str) -> ArrayOperation:
        if operation not in self._array_operations:
            self._probe_elementwise_operation(operation)
        return self._array_operations[operation]

    def lookup_function(self, operation: str) -> Callable:
        return self.lookup_array_operation(operation).operation_call
    
    def __repr__(self):
        """
        Implements the __repr__ method used in print.
        """
        return f"ArrayBackend(impl={self.impl})"
    
    def __eq__(self, other):
        """
        Implements the `==` operator.
        It compares if `other` is also an `ArrayBackend` and if `self` and `other` have the same implementation `impl`
        """
        return isinstance(other, ArrayBackend) and self.impl == other.impl

def lookup_array_backend(impl: str) -> ArrayBackend:
    """
    Convenience function for getting an `ArrayBackend` from an `impl` argument.
    This is helpful to both ensure that a backend actually exists and to retrieve it.

    Parameters
    ----------
    impl : str
        backend identifier

    Examples
    --------
    >>> lookup_array_backend('numpy')
    ArrayBackend(impl=numpy)
    >>> lookup_array_backend('something_else')
    Traceback (most recent call last):
    KeyError: "The implementation something_else is not supported by ODL. Please select a backend in ['numpy']"
    >>> lookup_array_backend(72)
    Traceback (most recent call last):
    AssertionError: f"The impl parameter must be a string, got int"
    """
    assert isinstance(impl, str), f"The impl parameter must be a string, got {type(impl)}"
    try:
        return _registered_array_backends[impl]
    except KeyError:
        if impl in standard_known_backends:
            raise KeyError(f"The implementation ‘{impl}’ is not available here, likely due to a missing package. Try installing {standard_known_backends[impl]} using pip / conda / uv.")
        else:
            raise KeyError(f"The implementation {impl} is not supported by ODL. Please select a backend in {_registered_array_backends.keys()}")

def get_array_and_backend(x, must_be_contiguous=False):
    """
    Convenience function for getting an `ArrayBackend` from an `array-like` argument. 

    Parameters
    ----------
    x : Array-Like.
        It can be a `np.ndarray`, a `torch.Tensor`, an ODL `Tensor` or a `ProductSpaceElement`. Object to return the `ArrayBackend` and actual underlying array from.
    must_be_contiguous : bool
        Boolean flag to indicate whether or not to make the array contiguous.

    Returns
    -------
    x : actual array 
        -> unwrapped from the LinearSpaceElement
        -> returned as is if it was already an array.
    backend : ODL `ArrayBackend` object

    Examples
    --------
    >>> array, backend = get_array_and_backend(np.zeros(2))
    >>> array
    array([ 0.,  0.])
    >>> backend
    ArrayBackend(impl=numpy)
    >>> array, backend = get_array_and_backend([1,2,3])
    Traceback (most recent call last):
    ValueError: f"The registered array backends are ['numpy']. The argument provided is a list, check that the backend you want to use is supported and has been correctly instanciated."
    """
    from odl.core.space.base_tensors import Tensor
    if isinstance(x, Tensor):
        return x.asarray(must_be_contiguous=must_be_contiguous), x.space.array_backend

    from odl.core.space.pspace import ProductSpaceElement
    if isinstance(x, ProductSpaceElement):
        return get_array_and_backend(x.asarray(), must_be_contiguous=must_be_contiguous)

    for backend in _registered_array_backends.values():
        backend : ArrayBackend
        if isinstance(x, backend.array_type) or x in backend.available_dtypes.values():
            if must_be_contiguous:
                return backend.make_contiguous(x), backend
            else:
                return x, backend

    else:
        raise ValueError(f"The registered array backends are {list(_registered_array_backends.keys())}. The argument provided is a {type(x)}, check that the backend you want to use is supported and has been correctly instanciated.")

def is_array_supported(x):
    for backend in _registered_array_backends.values():
        backend : ArrayBackend
        if isinstance(x, backend.array_type):
            return True
    return False

def check_device(impl:str, device: Union[str, object]) -> str:
    """
    Checks the device argument.
    This checks that the device requested is available and that its compatible with the backend requested.

    If successful, returns the standard string identifier of the device.

    Parameters
    ----------
    impl : str
        backend identifier
    device : str or backend-specific device-object
        Device identifier

    Examples
    --------
    >>> odl.check_device('numpy', 'cpu')
    'cpu'
    >>> odl.check_device('numpy', 'anything_but_cpu')
    Traceback (most recent call last):
    ...
    ValueError: For numpy Backend, only devices ['cpu'] are present, but anything_but_cpu was provided.
    """
    backend = lookup_array_backend(impl)
    for known_device in backend.available_devices:
        if device == known_device:
            return device
        elif str(device) == known_device:
            # This works at least for PyTorch, but it is not clear
            # how general this is.
            return str(device)

    raise ValueError(f"For {impl} Backend, only devices {backend.available_devices} are present, but {device} was provided.")
    
def _dtype_info(array_namespace, dtype):
    """
    Return min, max, and kind ('bool', 'int', 'uint', 'float') for a given dtype.
    Works across Array API backends.
    """
    name = str(dtype)
    if "bool" in name:
        return 0, 1, "bool"
    if "int" in name and not "uint" in name:
        iinfo = array_namespace.iinfo(dtype)
        return iinfo.min, iinfo.max, "int"
    if "uint" in name:
        iinfo = array_namespace.iinfo(dtype)
        return iinfo.min, iinfo.max, "uint"
    if "float" in name:
        finfo = array_namespace.finfo(dtype)
        # floats have no exact min/max, but finfo.min/max are usable for range checks
        return finfo.min, finfo.max, "float"
    raise ValueError(f"Unsupported dtype: {dtype}")

def can_cast(array_namespace, from_dtype, to_dtype, casting="safe"):
    """
    NumPy-like can_cast for Python Array API backends.
    Supports 'safe', 'same_kind', and 'unsafe' casting.
    """
    # Convert arrays to dtypes
    if hasattr(from_dtype, "dtype"):
        from_dtype = from_dtype.dtype
    if hasattr(to_dtype, "dtype"):
        to_dtype = to_dtype.dtype

    # Same type always allowed
    if from_dtype == to_dtype:
        return True

    # Unsafe allows anything
    if casting == "unsafe":
        return True

    # Determine type categories
    f_min, f_max, f_kind = _dtype_info(array_namespace, from_dtype)
    t_min, t_max, t_kind = _dtype_info(array_namespace, to_dtype)

    # Safe casting: all values of from_dtype must fit in to_dtype
    if casting == "safe":
        if f_kind == "bool":
            return True  # bool -> anything is safe
        if t_kind == "bool":
            return False  # non-bool -> bool is unsafe
        if f_kind in ("int", "uint") and t_kind in ("int", "uint", "float"):
            return f_min >= t_min and f_max <= t_max
        if f_kind == "float" and t_kind == "float":
            return array_namespace.finfo(to_dtype).precision >= array_namespace.finfo(from_dtype).precision
        return False

    # Same-kind casting: allow within same category or safe upcast to float
    if casting == "same_kind":
        if f_kind == t_kind:
            return True
        # int/uint to float is allowed if range fits
        if f_kind in ("int", "uint") and t_kind == "float":
            return f_min >= t_min and f_max <= t_max
        return False

    raise ValueError(f"Unsupported casting rule: {casting}")

if __name__ =='__main__':
    check_device('numpy', 'cpu')



import pytest

import odl

from odl.core.array_API_support import odl_all_equal

from odl.core.util.pytest_config import IMPL_DEVICE_PAIRS
from odl.core.util.testutils import (
    noise_elements, simple_fixture)

DEFAULT_SHAPE = (4,4)

DEFAULT_FILL = 5

from_array = simple_fixture(
    'from_array', ["asarray", "empty_like", "full_like", 'ones_like', 'tril', 'triu', 'zeros_like']
    )

from_impl = simple_fixture(
    'from_impl', ['arange', 'empty', 'eye', "full", 'linspace', 'meshgrid', 'ones', 'zeros']
    )


@pytest.fixture(scope='module', params=IMPL_DEVICE_PAIRS)
def float_tspace(request, odl_real_floating_dtype):
    impl, device = request.param
    return odl.tensor_space(
        shape=DEFAULT_SHAPE, 
        dtype=odl_real_floating_dtype, 
        impl=impl, 
        device=device
    )

def test_from_array(float_tspace, from_array):
    ns = float_tspace.array_namespace
    arr_fn = getattr(ns, from_array)
    odl_fn = getattr(odl, from_array)

    x_arr, x = noise_elements(float_tspace, 1)

    if from_array == 'full_like':
        y_arr = arr_fn(x_arr, fill_value=DEFAULT_FILL)
        y     = odl_fn(x, fill_value=DEFAULT_FILL)
    else:
        y_arr = arr_fn(x_arr)
        y     = odl_fn(x)
    if from_array == 'empty_like':
        pytest.skip("Skipping equality check for empty_like")
    
    else:
        assert odl_all_equal(y_arr, y)

# Pytorch and Numpy API still vary, making the systematic testing of these functions premature
# def test_from_impl(float_tspace, from_impl):
#     ns = float_tspace.array_namespace
#     arr_fn = getattr(ns, from_impl)
#     odl_fn = getattr(odl, from_impl)

#     # x_arr, x = noise_elements(float_tspace, 1)
#     args = ()
#     kwargs = {
#             'shape'  : (4,4),
#             'dtype'  : float_tspace.dtype_identifier,
#             'device' : float_tspace.device
#         }
#     if from_impl == 'arange':
#         args = [1]
#         kwargs['start'] = 1
#         kwargs['stop'] = 10
#         kwargs['step'] = 1
            
#     elif from_impl == 'eye':
#         kwargs['n_rows'] = 4
#         kwargs['n_cols'] = 4
#         kwargs['k'] = 0

#     elif from_impl == 'meshgrid':
#         args = [
#             float_tspace.array_backend.array_constructor([0,1,2,3], 
#             device = float_tspace.device,
#             dtype = float_tspace.dtype),
#             float_tspace.array_backend.array_constructor([0,1,2,3], 
#             device = float_tspace.device,
#             dtype = float_tspace.dtype)
#             ]

#     elif from_impl == 'tril' or from_impl == 'triu':
#         kwargs['k'] = 2

#     print(args, kwargs)
#     assert odl_all_equal(
#         arr_fn(*args, **kwargs), odl_fn(*args, **kwargs)
    # )



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Unit tests for `DiscretizedSpace`."""

from __future__ import division

import numpy as np

import odl
import pytest
from odl.core.discr.discr_space import DiscretizedSpace, DiscretizedSpaceElement
from odl.core.space.base_tensors import TensorSpace, default_dtype
from odl.backends.arrays.npy_tensors import NumpyTensor
from odl.core.util.dtype_utils import COMPLEX_DTYPES, DTYPE_SHORTHANDS
from odl.core.util.testutils import (
    all_almost_equal, all_equal, noise_elements, simple_fixture, default_precision_dict)
from odl.core.array_API_support import lookup_array_backend
# --- Pytest fixtures --- #


exponent = simple_fixture('exponent', [2.0, 1.0, float('inf'), 0.5, 1.5])
power = simple_fixture('power', [1.0, 2.0, 0.5, -0.5, -1.0, -2.0])
shape = simple_fixture('shape', [(2, 3, 4), (3, 4), (2,), (1,), (1, 1, 1)])
power = simple_fixture('power', [1.0, 2.0, 0.5, -0.5, -1.0, -2.0])


# --- DiscretizedSpace --- #


def test_discretizedspace_init(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Test initialization and basic properties of DiscretizedSpace."""
    # Real space
    part = odl.uniform_partition([0, 0], [1, 1], (2, 4))
    tspace = odl.rn(part.shape, impl=impl, device=device)

    discr = DiscretizedSpace(part, tspace)
    assert discr.tspace == tspace
    assert discr.partition == part
    assert discr.exponent == tspace.exponent
    assert discr.axis_labels == ('$x$', '$y$')
    assert discr.is_real

    # Complex space
    tspace_c = odl.cn(part.shape, impl=impl, device=device)
    discr = DiscretizedSpace(part, tspace_c)
    assert discr.is_complex

    # Make sure repr shows something
    assert repr(discr) != ''

    # Error scenarios
    part_1d = odl.uniform_partition(0, 1, 2)
    with pytest.raises(ValueError):
        DiscretizedSpace(part_1d, tspace)  # wrong dimensionality

    part_diffshp = odl.uniform_partition([0, 0], [1, 1], (3, 4))
    with pytest.raises(ValueError):
        DiscretizedSpace(part_diffshp, tspace)  # shape mismatch


def test_empty(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Check if empty spaces behave as expected and all methods work."""
    discr = odl.uniform_discr([], [], (), impl=impl, device=device)

    assert discr.axis_labels == ()
    assert discr.tangent_bundle == odl.ProductSpace(field=odl.RealNumbers())
    assert discr.complex_space == odl.uniform_discr([], [], (), dtype=complex, impl=impl, device=device)
    hash(discr)
    assert repr(discr) != ''

    elem = discr.element(1.0)
    assert all_equal(elem.asarray(), 1.0)
    assert all_equal(elem.real, 1.0)
    assert all_equal(elem.imag, 0.0)
    assert all_equal(elem.conj(), 1.0)


# --- uniform_discr --- #


def test_factory_dtypes(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Check dtypes of spaces from factory function."""
    real_float_dtypes = ["float32", "float64"]
    nonfloat_dtypes = ["int8", "int16", "int32", "int64",
                       "uint8", "uint16", "uint32", "uint64"]
    complex_float_dtypes = ["complex64", "complex128"]

    for dtype in real_float_dtypes:
        try:
            discr = odl.uniform_discr(0, 1, 10, impl=impl, dtype=dtype, device=device)
        except TypeError:
            continue
        else:
            assert isinstance(discr.tspace, TensorSpace)
            assert discr.tspace.impl == impl
            assert discr.is_real

    for dtype in nonfloat_dtypes:
        try:
            discr = odl.uniform_discr(0, 1, 10, impl=impl, dtype=dtype, device=device)
        except TypeError:
            continue
        else:
            assert isinstance(discr.tspace, TensorSpace)
            assert discr.tspace.impl == impl
            assert discr.tspace.element().space.dtype_identifier == dtype

    for dtype in complex_float_dtypes:
        try:
            discr = odl.uniform_discr(0, 1, 10, impl=impl, dtype=dtype, device=device)
        except TypeError:
            continue
        else:
            assert isinstance(discr.tspace, TensorSpace)
            assert discr.tspace.impl == impl
            assert discr.is_complex
            assert discr.tspace.element().space.dtype_identifier == dtype


def test_uniform_discr_init_real(odl_impl_device_pairs):
    """Test initialization and basic properties with uniform_discr, real."""
    impl, device = odl_impl_device_pairs

    # 1D
    discr = odl.uniform_discr(0, 1, 10, impl=impl, device=device)
    assert isinstance(discr, DiscretizedSpace)
    assert isinstance(discr.tspace, TensorSpace)
    assert discr.impl == impl
    assert discr.is_real
    assert discr.tspace.exponent == 2.0
    assert discr.dtype == default_dtype(impl, field=odl.RealNumbers())
    assert discr.is_real
    assert not discr.is_complex
    assert all_equal(discr.min_pt, [0])
    assert all_equal(discr.max_pt, [1])
    assert discr.shape == (10,)
    assert repr(discr)

    discr = odl.uniform_discr(0, 1, 10, impl=impl, exponent=1.0, device=device)
    assert discr.exponent == 1.0

    # 2D
    discr = odl.uniform_discr([0, 0], [1, 1], (5, 5), impl=impl, device=device)
    assert all_equal(discr.min_pt, np.array([0, 0]))
    assert all_equal(discr.max_pt, np.array([1, 1]))
    assert discr.shape == (5, 5)

    # nd
    discr = odl.uniform_discr([0] * 10, [1] * 10, (5,) * 10, impl=impl, device=device)
    assert all_equal(discr.min_pt, np.zeros(10))
    assert all_equal(discr.max_pt, np.ones(10))
    assert discr.shape == (5,) * 10


# ## Why does this test fail if impl != numpy?
# def test_uniform_discr_init_complex(odl_tspace_impl):
#     """Test initialization and basic properties with uniform_discr, complex."""
#     impl = odl_tspace_impl
#     if impl != 'numpy':
#         pytest.xfail(reason='complex dtypes not supported')

#     discr = odl.uniform_discr(0, 1, 10, dtype='complex', impl=impl)
#     assert discr.is_complex
#     assert discr.dtype == default_dtype(impl, field=odl.ComplexNumbers())


# --- DiscretizedSpace methods --- #


def test_discretizedspace_element(odl_impl_device_pairs):
    """Test creation and membership of DiscretizedSpace elements."""
    impl, device = odl_impl_device_pairs
    # Creation from scratch
    # 1D
    discr = odl.uniform_discr(0, 1, 3, impl=impl, device=device)
    weight = 1.0 if exponent == float('inf') else discr.cell_volume
    tspace = odl.rn(3, weighting=weight, impl=impl, device=device)
    elem = discr.element()
    assert elem in discr
    assert elem.tensor in tspace

    # 2D
    discr = odl.uniform_discr([0, 0], [1, 1], (3, 3), impl=impl, device=device)
    weight = 1.0 if exponent == float('inf') else discr.cell_volume
    tspace = odl.rn((3, 3), weighting=weight, impl=impl, device=device)
    elem = discr.element()
    assert elem in discr
    assert elem.tensor in tspace


def test_discretizedspace_element_from_array(odl_impl_device_pairs):
    """Test creation of DiscretizedSpace elements from arrays."""
    impl, device = odl_impl_device_pairs
    # 1D
    discr = odl.uniform_discr(0, 1, 3, impl=impl, device=device)
    elem = discr.element([1, 2, 3])
    assert all_equal(elem.tensor, [1, 2, 3])

    assert isinstance(elem, DiscretizedSpaceElement)
    assert isinstance(elem.tensor, discr.tspace.element_type)
    assert all_equal(elem.tensor, [1, 2, 3])

def test_element_from_array_2d(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    discr = odl.uniform_discr([0, 0], [1, 1], [2, 2], impl=impl, device=device)
    elem = discr.element([[1, 2],
                          [3, 4]])

    assert isinstance(elem, DiscretizedSpaceElement)
    if impl=='numpy':
        assert isinstance(elem.tensor, NumpyTensor)
    assert all_equal(elem, [[1, 2],
                            [3, 4]])

    with pytest.raises(ValueError):
        discr.element([1, 2, 3])  # wrong size & shape
    with pytest.raises(ValueError):
        discr.element([1, 2, 3, 4])  # wrong shape
    with pytest.raises(ValueError):
        discr.element([[1],
                       [2],
                       [3],
                       [4]])  # wrong shape


def test_element_from_function_1d(odl_impl_device_pairs):
    """Test creation of DiscretizedSpace elements from functions in 1D."""
    impl, device = odl_impl_device_pairs
    space = odl.uniform_discr(-1, 1, 4, impl=impl, device=device)
    points = space.points().squeeze()

    backend = lookup_array_backend(impl)
    namespace = backend.array_namespace 
    # Without parameter
    def f(x):
        zero = namespace.zeros_like(x)
        return x * 2 + namespace.maximum(x, zero)

    elem_f = space.element(f)
    true_elem = [x * 2 + max(x, 0) for x in points]
    assert all_equal(elem_f, true_elem)

    # Without parameter, using same syntax as in higher dimensions
    def f(x):
        zero = namespace.zeros_like(x[0])
        return x[0] * 2 + namespace.maximum(x[0], zero)

    elem_f = space.element(f)
    true_elem = [x * 2 + max(x, 0) for x in points]
    assert all_equal(elem_f, true_elem)

    # With parameter
    def f(x, **kwargs):
        c = kwargs.pop('c', 0)
        zero = namespace.zeros_like(x)
        return x * c + namespace.maximum(x, zero)

    elem_f_default = space.element(f)
    true_elem = [x * 0 + max(x, 0) for x in points]
    assert all_equal(elem_f_default, true_elem)

    elem_f_2 = space.element(f, c=2)
    true_elem = [x * 2 + max(x, 0) for x in points]
    assert all_equal(elem_f_2, true_elem)

    # Using a lambda
    elem_lam = space.element(lambda x: -x ** 2)
    true_elem = [-x ** 2 for x in points]
    assert all_equal(elem_lam, true_elem)

    # Broadcast from constant function
    elem_lam = space.element(lambda x: 1.0)
    true_elem = [1.0 for _ in points]
    assert all_equal(elem_lam, true_elem)


def test_element_from_function_2d(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Test creation of DiscretizedSpace elements from functions in 2D."""
    space = odl.uniform_discr([-1, -1], [1, 1], (2, 3), impl=impl, device=device)
    points = space.points()

    backend = lookup_array_backend(impl)
    namespace = backend.array_namespace 

    # Without parameter
    def f(x):
        zero = namespace.zeros_like(x[0])
        return x[0] ** 2 + namespace.maximum(x[1], zero)

    elem_f = space.element(f)
    true_elem = np.reshape(
        [x[0] ** 2 + max(x[1], 0) for x in points], space.shape
    )
    assert all_equal(elem_f, true_elem)

    # With parameter
    def f(x, **kwargs):
        c = kwargs.pop('c', 0) * namespace.ones_like(x[0])
        return x[0] ** 2 + namespace.maximum(x[1], c)

    elem_f_default = space.element(f)
    true_elem = np.reshape(
        [x[0] ** 2 + max(x[1], 0) for x in points], space.shape
    )
    assert all_equal(elem_f_default, true_elem)

    elem_f_2 = space.element(f, c=1)
    true_elem = np.reshape(
        [x[0] ** 2 + max(x[1], 1) for x in points], space.shape
    )
    assert all_equal(elem_f_2, true_elem)

    # Using a lambda
    elem_lam = space.element(lambda x: x[0] - x[1])
    true_elem = np.reshape([x[0] - x[1] for x in points], space.shape)
    assert all_equal(elem_lam, true_elem)

    # Using broadcasting
    elem_lam = space.element(lambda x: x[0])
    true_elem = np.reshape([x[0] for x in points], space.shape)
    assert all_equal(elem_lam, true_elem)

    elem_lam = space.element(lambda x: x[1])
    true_elem = np.reshape([x[1] for x in points], space.shape)
    assert all_equal(elem_lam, true_elem)

    # Broadcast from constant function
    elem_lam = space.element(lambda x: 1.0)
    true_elem = np.reshape([1.0 for _ in points], space.shape)
    assert all_equal(elem_lam, true_elem)


def test_discretizedspace_zero_one(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Test the zero and one element creators of DiscretizedSpace."""
    discr = odl.uniform_discr(0, 1, 3, impl=impl, device=device)

    zero = discr.zero()
    assert zero in discr
    assert all_equal(zero, [0, 0, 0])

    one = discr.one()
    assert one in discr
    assert all_equal(one, [1, 1, 1])


def test_equals_space(exponent, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    x1 = odl.uniform_discr(0, 1, 3, exponent=exponent, impl=impl, device=device)
    x2 = odl.uniform_discr(0, 1, 3, exponent=exponent, impl=impl, device=device)
    y = odl.uniform_discr(0, 1, 4, exponent=exponent, impl=impl, device=device)

    assert x1 is x1
    assert x1 is not x2
    assert x1 is not y
    assert x1 == x1
    assert x1 == x2
    assert x1 != y
    assert hash(x1) == hash(x2)
    assert hash(x1) != hash(y)


def test_equals_vec(exponent, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    discr = odl.uniform_discr(0, 1, 3, exponent=exponent, impl=impl, device=device)
    discr2 = odl.uniform_discr(0, 1, 4, exponent=exponent, impl=impl, device=device)
    x1 = discr.element([1, 2, 3])
    x2 = discr.element([1, 2, 3])
    y = discr.element([2, 2, 3])
    z = discr2.element([1, 2, 3, 4])

    assert x1 is x1
    assert x1 is not x2
    assert x1 is not y
    assert x1 == x1
    assert x1 == x2
    assert x1 != y
    assert x1 != z


def _test_unary_operator(discr, function):
    # Verify that the statement y=function(x) gives equivalent results
    # to NumPy
    x_arr, x = noise_elements(discr)
    y_arr = function(x_arr)
    y = function(x)
    assert all_almost_equal([x, y], [x_arr, y_arr])


def _test_binary_operator(discr, function):
    # Verify that the statement z=function(x,y) gives equivalent results
    # to NumPy
    [x_arr, y_arr], [x, y] = noise_elements(discr, 2)
    z_arr = function(x_arr, y_arr)
    z = function(x, y)
    assert all_almost_equal([x, y, z], [x_arr, y_arr, z_arr])


def test_operators(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # Test of all operator overloads against the corresponding NumPy
    # implementation
    discr = odl.uniform_discr(0, 1, 10, impl=impl, device=device)

    # Unary operators
    _test_unary_operator(discr, lambda x: +x)
    _test_unary_operator(discr, lambda x: -x)

    # Scalar addition
    for scalar in [-31.2, -1, 0, 1, 2.13]:
        def iadd(x):
            x += scalar
        _test_unary_operator(discr, iadd)
        _test_unary_operator(discr, lambda x: x + scalar)

    # Scalar subtraction
    for scalar in [-31.2, -1, 0, 1, 2.13]:
        def isub(x):
            x -= scalar
        _test_unary_operator(discr, isub)
        _test_unary_operator(discr, lambda x: x - scalar)

    # Scalar multiplication
    for scalar in [-31.2, -1, 0, 1, 2.13]:
        def imul(x):
            x *= scalar
        _test_unary_operator(discr, imul)
        _test_unary_operator(discr, lambda x: x * scalar)

    # Scalar division
    for scalar in [-31.2, -1, 1, 2.13]:
        def idiv(x):
            x /= scalar
        _test_unary_operator(discr, idiv)
        _test_unary_operator(discr, lambda x: x / scalar)

    # Incremental operations
    def iadd(x, y):
        x += y

    def isub(x, y):
        x -= y

    def imul(x, y):
        x *= y

    def idiv(x, y):
        x /= y

    _test_binary_operator(discr, iadd)
    _test_binary_operator(discr, isub)
    _test_binary_operator(discr, imul)
    _test_binary_operator(discr, idiv)

    # Incremental operators with aliased inputs
    def iadd_aliased(x):
        x += x

    def isub_aliased(x):
        x -= x

    def imul_aliased(x):
        x *= x

    def idiv_aliased(x):
        x /= x

    _test_unary_operator(discr, iadd_aliased)
    _test_unary_operator(discr, isub_aliased)
    _test_unary_operator(discr, imul_aliased)
    _test_unary_operator(discr, idiv_aliased)

    # Binary operators
    _test_binary_operator(discr, lambda x, y: x + y)
    _test_binary_operator(discr, lambda x, y: x - y)
    _test_binary_operator(discr, lambda x, y: x * y)
    _test_binary_operator(discr, lambda x, y: x / y)

    # Binary with aliased inputs
    _test_unary_operator(discr, lambda x: x + x)
    _test_unary_operator(discr, lambda x: x - x)
    _test_unary_operator(discr, lambda x: x * x)
    _test_unary_operator(discr, lambda x: x / x)


def test_getitem(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    discr = odl.uniform_discr(0, 1, 3, impl=impl, device=device)
    elem = discr.element([1, 2, 3])

    assert all_equal(elem, [1, 2, 3])


def test_getslice(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    discr = odl.uniform_discr(0, 1, 3, impl=impl, device=device)
    elem = discr.element([1, 2, 3])
    tspace_impl = discr.tspace.element_type
    assert isinstance(elem[:], tspace_impl)
    assert all_equal(elem[:], [1, 2, 3])

    discr = odl.uniform_discr(0, 1, 3, dtype=complex)
    tspace_impl = discr.tspace.element_type
    elem = discr.element([1 + 2j, 2 - 2j, 3])

    assert isinstance(elem[:], tspace_impl)
    assert all_equal(elem[:], [1 + 2j, 2 - 2j, 3])


def test_setitem(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    discr = odl.uniform_discr(0, 1, 3, impl=impl, device=device)
    elem = discr.element([1, 2, 3])
    elem[0] = 4
    elem[1] = 5
    elem[2] = 6

    assert all_equal(elem, [4, 5, 6])


def test_setitem_nd(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs

    # 1D
    discr = odl.uniform_discr(0, 1, 3, impl=impl, device=device)
    elem = discr.element([1, 2, 3])

    backend = discr.array_backend

    elem[:] = [4, 5, 6]
    assert all_equal(elem, [4, 5, 6])

    elem[:] = backend.array_constructor([3, 2, 1], device=device)
    assert all_equal(elem, [3, 2, 1])

    elem[:] = 0
    assert all_equal(elem, [0, 0, 0])

    elem[:] = [1]
    assert all_equal(elem, [1, 1, 1])

    error = ValueError if impl =='numpy' else RuntimeError
    with pytest.raises(error):
        elem[:] = [0, 0]  # bad shape

    with pytest.raises(error):
        elem[:] = [0, 0, 1, 2]  # bad shape

    # 2D
    discr = odl.uniform_discr([0, 0], [1, 1], [3, 2])

    elem = discr.element([[1, 2],
                          [3, 4],
                          [5, 6]])

    elem[:] = [[-1, -2],
               [-3, -4],
               [-5, -6]]
    assert all_equal(elem, [[-1, -2],
                            [-3, -4],
                            [-5, -6]])

    # arr = np.arange(6, 12).reshape([3, 2])
    arr = odl.arange(impl=impl, start=6, stop=12).reshape([3, 2])
    elem[:] = arr
    assert all_equal(elem, arr)

    elem[:] = 0
    assert all_equal(elem, odl.zeros(impl=impl, shape=elem.shape))

    elem[:] = [1]
    assert all_equal(elem, odl.ones(impl=impl, shape=elem.shape))

    elem[:] = [0, 0]  # broadcasting assignment
    assert all_equal(elem,odl.zeros(impl=impl, shape=elem.shape))

    with pytest.raises(ValueError):
        elem[:] = [0, 0, 0]  # bad shape

    with pytest.raises(ValueError):
        elem[:] = odl.arange(impl=impl, start=6)  # bad shape (6,)

    with pytest.raises(ValueError):
        elem[:] = odl.ones(impl=impl, shape=(2, 3))[..., None]  # bad shape (2, 3, 1)

    with pytest.raises(ValueError):
        arr = odl.arange(impl=impl, start=6, stop=12).reshape([3, 2])
        elem[:] = arr.T  # bad shape (2, 3)

    # nD
    shape = (3,) * 3 + (4,) * 3
    discr = odl.uniform_discr([0] * 6, [1] * 6, shape, impl=impl, device=device)
    size = np.prod(shape)
    elem = discr.element(np.zeros(shape))

    arr = odl.arange(impl=impl, start=size).reshape(shape)

    elem[:] = arr
    assert all_equal(elem, arr)

    elem[:] = 0
    assert all_equal(elem, odl.zeros(impl=impl, shape=elem.shape))

    elem[:] = [1]
    assert all_equal(elem, odl.ones(impl=impl, shape=elem.shape))

    error = ValueError if impl =='numpy' else RuntimeError
    with pytest.raises(error):
        # Reversed shape -> bad
        elem[:] = odl.arange(impl=impl, start=size).reshape((4,) * 3 + (3,) * 3)


def test_setslice(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    discr = odl.uniform_discr(0, 1, 3, impl=impl, device=device)
    elem = discr.element([1, 2, 3])

    elem[:] = [4, 5, 6]
    assert all_equal(elem, [4, 5, 6])


def test_asarray_2d(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Test the asarray method."""
    discr = odl.uniform_discr([0, 0], [1, 1], [2, 2], impl=impl, device=device)
    elem = discr.element([[1, 2],
                          [3, 4]])

    arr = elem.asarray()
    assert all_equal(arr, [[1, 2],
                           [3, 4]])
    
    # test out parameter
    out_c = odl.empty(impl=impl, shape=[2, 2])
    result_c = elem.asarray(out=out_c)
    assert result_c is out_c
    assert all_equal(out_c, [[1, 2],
                             [3, 4]])
    # Try wrong shape
    out_wrong_shape = odl.empty(impl=impl, shape=[2, 3])
    error = ValueError if impl =='numpy' else RuntimeError
    with pytest.raises(error):
        elem.asarray(out=out_wrong_shape)


def test_transpose(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    discr = odl.uniform_discr([0, 0], [1, 1], [2, 2], impl=impl, device=device)
    x = discr.element([[1, 2], [3, 4]])
    y = discr.element([[5, 6], [7, 8]])

    assert isinstance(x.T, odl.Operator)
    assert x.T.is_linear

    assert x.T(y) == x.inner(y)
    assert x.T.T == x
    assert all_equal(x.T.adjoint(1.0), x)


def test_cell_sides(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # Non-degenerated case, should be same as cell size
    discr = odl.uniform_discr([0, 0], [1, 1], [2, 2], impl=impl, device=device)
    elem = discr.element()

    assert all_equal(discr.cell_sides, [0.5] * 2)
    assert all_equal(elem.cell_sides, [0.5] * 2)

    # Degenerated case, uses interval size in 1-point dimensions
    discr = odl.uniform_discr([0, 0], [1, 1], [2, 1], impl=impl, device=device)
    elem = discr.element()

    assert all_equal(discr.cell_sides, [0.5, 1])
    assert all_equal(elem.cell_sides, [0.5, 1])


def test_cell_volume(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # Non-degenerated case
    discr = odl.uniform_discr([0, 0], [1, 1], [2, 2], impl=impl, device=device)
    elem = discr.element()

    assert discr.cell_volume == 0.25
    assert elem.cell_volume == 0.25

    # Degenerated case, uses interval size in 1-point dimensions
    discr = odl.uniform_discr([0, 0], [1, 1], [2, 1])
    elem = discr.element()

    assert discr.cell_volume == 0.5
    assert elem.cell_volume == 0.5


def test_astype(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    rdiscr = odl.uniform_discr([0, 0], [1, 1], [2, 2], dtype='float64', impl=impl, device=device)
    cdiscr = odl.uniform_discr([0, 0], [1, 1], [2, 2], dtype='complex128', impl=impl, device=device)
    rdiscr_s = odl.uniform_discr([0, 0], [1, 1], [2, 2], dtype='float32', impl=impl, device=device)
    cdiscr_s = odl.uniform_discr([0, 0], [1, 1], [2, 2], dtype='complex64', impl=impl, device=device)

    # Real
    assert rdiscr.astype('float32') == rdiscr_s
    assert rdiscr.astype('float64') is rdiscr
    assert rdiscr.real_space is rdiscr
    assert rdiscr.astype('complex64') == cdiscr_s
    assert rdiscr.astype('complex128') == cdiscr
    assert rdiscr.complex_space == cdiscr

    # Complex
    assert cdiscr.astype('complex64') == cdiscr_s
    assert cdiscr.astype('complex128') is cdiscr
    assert cdiscr.complex_space is cdiscr
    assert cdiscr.astype('float32') == rdiscr_s
    assert cdiscr.astype('float64') == rdiscr
    assert cdiscr.real_space == rdiscr

    # More exotic dtype
    discr = odl.uniform_discr([0, 0], [1, 1], [2, 2], dtype=bool, impl=impl, device=device)
    as_float = discr.astype(float)
    assert as_float.dtype_identifier == DTYPE_SHORTHANDS[float]
    assert not as_float.is_weighted
    as_complex = discr.astype(complex)
    assert as_complex.dtype_identifier == DTYPE_SHORTHANDS[complex]
    assert not as_complex.is_weighted



def test_real_imag(odl_elem_order, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Check if real and imaginary parts can be read and written to."""
    order = odl_elem_order
    tspace_cls = odl.core.space.entry_points.tensor_space_impl(impl)
    for dtype in COMPLEX_DTYPES:
        cdiscr = odl.uniform_discr([0, 0], [1, 1], [2, 2], dtype=dtype, impl=impl, device=device)
        rdiscr = cdiscr.real_space

        # Get real and imag
        x = cdiscr.element([[1 - 1j, 2 - 2j],
                            [3 - 3j, 4 - 4j]])
        assert x.real in rdiscr
        assert all_equal(x.real, [[1, 2],
                                  [3, 4]])
        assert x.imag in rdiscr
        assert all_equal(x.imag, [[-1, -2],
                                  [-3, -4]])

        # Set with different data types and shapes
        for assigntype in [ lambda x: x, tuple, rdiscr.element ]:


            # Using setters
            x = cdiscr.zero()
            new_real = assigntype([[2, 3],
                                 [4, 5]])
            x.real = new_real
            assert all_equal(x.real, [[2, 3],
                                      [4, 5]])

            x = cdiscr.zero()
            x.imag = assigntype([[4, 5],
                                 [6, 7]])
            assert all_equal(x.imag, [[4, 5],
                                      [6, 7]])

            # With [:] assignment
            x = cdiscr.zero()
            x.real[:] = assigntype([[2, 3],
                                    [4, 5]])
            assert all_equal(x.real, [[2, 3],
                                      [4, 5]])

            x = cdiscr.zero()
            x.imag[:] = assigntype([[2, 3],
                                    [4, 5]])
            assert all_equal(x.imag, [[2, 3],
                                      [4, 5]])

        # Setting with scalars
        x = cdiscr.zero()
        x.real = 1
        assert all_equal(x.real, [[1, 1],
                                  [1, 1]])

        x = cdiscr.zero()
        x.imag = -1
        assert all_equal(x.imag, [[-1, -1],
                                  [-1, -1]])

    # Incompatible shapes
    with pytest.raises(ValueError):
        x.real = [4, 5, 6, 7]
    with pytest.raises(ValueError):
        x.imag = [4, 5, 6, 7]


def test_reduction(odl_reduction, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    name = odl_reduction
    space = odl.uniform_discr([0, 0], [1, 1], [2, 2], impl=impl, device=device)

    reduction = getattr(odl, name)
    backend_reduction = getattr(space.array_namespace, name)

    # Create some data
    x_arr, x = noise_elements(space, 1)
    arr_red = space.array_backend.to_cpu(backend_reduction(x_arr))
    odl_red = space.array_backend.to_cpu(reduction(x))
    assert arr_red == pytest.approx(odl_red)


def test_power(power, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    space = odl.uniform_discr([0, 0], [1, 1], [2, 2], impl=impl, device=device)
    ns = space.array_namespace
    x_arr, x = noise_elements(space, 1)
    x_pos_arr = ns.abs(x_arr)
    x_neg_arr = -x_pos_arr
    x_pos = odl.abs(x)
    x_neg = -x_pos

    power_keyword = 'power' if impl == 'numpy' else 'pow'
    power_function = getattr(ns, power_keyword)

    if int(power) != power:
        # Make input positive to get real result
        for y in [x_pos_arr, x_neg_arr, x_pos, x_neg]:
            y += 0.1

    with np.errstate(invalid='ignore'):
        true_pos_pow = power_function(x_pos_arr, power)
        true_neg_pow = power_function(x_neg_arr, power)

    if int(power) != power and impl == 'cuda':
        with pytest.raises(ValueError):
            x_pos ** power
        with pytest.raises(ValueError):
            x_pos **= power
    else:
        with np.errstate(invalid='ignore'):
            assert all_almost_equal(x_pos ** power, true_pos_pow)
            if int(power) == power:
                assert all_almost_equal(x_neg ** power, true_neg_pow)

            x_pos **= power
            assert all_almost_equal(x_pos, true_pos_pow)

            if int(power) == power:
                x_neg **= power
                assert all_almost_equal(x_neg, true_neg_pow)


def test_inner_nonuniform(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Check if inner products are correct in non-uniform discretizations."""
    part = odl.nonuniform_partition([0, 2, 3, 5], min_pt=0, max_pt=5)
    backend = lookup_array_backend(impl)
    weights = backend.array_constructor(part.cell_sizes_vecs[0], device=device)
    tspace = odl.rn(part.size, weighting=weights, impl=impl, device=device)
    discr = odl.DiscretizedSpace(part, tspace)

    one = discr.one()
    linear = discr.element(lambda x: x)

    # Exact inner product is the integral from 0 to 5 of x, which is 5**2 / 2
    exact_inner = 5 ** 2 / 2.0
    inner = one.inner(linear)
    assert inner == pytest.approx(exact_inner)


def test_norm_nonuniform(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Check if norms are correct in non-uniform discretizations."""
    part = odl.nonuniform_partition([0, 2, 3, 5], min_pt=0, max_pt=5)

    backend = lookup_array_backend(impl)
    weights = backend.array_constructor(part.cell_sizes_vecs[0], device=device)
    
    tspace = odl.rn(part.size, weighting=weights, impl=impl, device=device)
    discr = odl.DiscretizedSpace(part, tspace)

    sqrt = discr.element(lambda x: backend.array_namespace.sqrt(x))

    # Exact norm is the square root of the integral from 0 to 5 of x,
    # which is sqrt(5**2 / 2)
    exact_norm = np.sqrt(5 ** 2 / 2.0)
    norm = sqrt.norm()
    assert norm == pytest.approx(exact_norm)


def test_norm_interval(exponent, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # Test the function f(x) = x^2 on the interval (0, 1). Its
    # L^p-norm is (1 + 2*p)^(-1/p) for finite p and 1 for p=inf
    p = exponent
    discr = odl.uniform_discr(0, 1, 10, exponent=p, impl=impl, device=device)

    func = discr.element(lambda x: x ** 2)
    if p == float('inf'):
        assert func.norm() <= 1  # Max at boundary not hit
    else:
        true_norm = (1 + 2 * p) ** (-1 / p)
        assert func.norm() == pytest.approx(true_norm, rel=1e-2)


def test_norm_rectangle(exponent, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # Test the function f(x) = x_0^2 * x_1^3 on (0, 1) x (-1, 1). Its
    # L^p-norm is ((1 + 2*p) * (1 + 3 * p) / 2)^(-1/p) for finite p
    # and 1 for p=inf
    p = exponent
    discr = odl.uniform_discr([0, -1], [1, 1], (20, 30), exponent=p, impl=impl, device=device)

    func = discr.element(lambda x: x[0] ** 2 * x[1] ** 3)
    if p == float('inf'):
        assert func.norm() <= 1  # Max at boundary not hit
    else:
        true_norm = ((1 + 2 * p) * (1 + 3 * p) / 2) ** (-1 / p)
        assert func.norm() == pytest.approx(true_norm, rel=1e-2)


def test_norm_rectangle_boundary(exponent, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # Check the constant function 1 in different situations regarding the
    # placement of the outermost grid points.
    dtype = 'float32'

    # Standard case
    discr = odl.uniform_discr(
        [-1, -2], [1, 2], (4, 8), dtype=dtype, impl=impl, device=device, exponent=exponent
    )
    if exponent == float('inf'):
        assert discr.one().norm() == 1
    else:
        assert (
            discr.one().norm()
            == pytest.approx(discr.domain.volume ** (1 / exponent))
        )

    # Nodes on the boundary (everywhere)
    discr = odl.uniform_discr(
        [-1, -2], [1, 2], (4, 8), dtype=dtype, impl=impl, device=device, exponent=exponent,
        nodes_on_bdry=True
    )
    if exponent == float('inf'):
        assert discr.one().norm() == 1
    else:
        assert (
            discr.one().norm()
            == pytest.approx(discr.domain.volume ** (1 / exponent))
        )

    # Nodes on the boundary (selective)
    discr = odl.uniform_discr(
        [-1, -2], [1, 2], (4, 8), dtype=dtype, impl=impl, device=device, exponent=exponent,
        nodes_on_bdry=((False, True), False)
    )
    if exponent == float('inf'):
        assert discr.one().norm() == 1
    else:
        assert (
            discr.one().norm()
            == pytest.approx(discr.domain.volume ** (1 / exponent))
        )

    discr = odl.uniform_discr(
        [-1, -2], [1, 2], (4, 8), dtype=dtype, impl=impl, device=device, exponent=exponent,
        nodes_on_bdry=(False, (True, False))
    )
    if exponent == float('inf'):
        assert discr.one().norm() == 1
    else:
        assert (
            discr.one().norm()
            == pytest.approx(discr.domain.volume ** (1 / exponent))
        )

    # Completely arbitrary boundary
    part = odl.RectPartition(
        odl.IntervalProd([-1, -2], [1, 2]),
        odl.uniform_grid([0, 0], [1, 1], (4, 4))
    )
    weight = 1.0 if exponent == float('inf') else part.cell_volume
    tspace = odl.rn(part.shape, dtype=dtype, impl=impl,
                    exponent=exponent, weighting=weight, device=device)
    discr = DiscretizedSpace(part, tspace)

    if exponent == float('inf'):
        assert discr.one().norm() == 1
    else:
        assert (
            discr.one().norm()
            == pytest.approx(discr.domain.volume ** (1 / exponent))
        )


def test_uniform_discr_fromdiscr_one_attr(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # Change 1 attribute
    discr = odl.uniform_discr([0, -1], [1, 1], [10, 5], impl=impl, device=device)
    # csides = [0.1, 0.4]

    # min_pt -> translate, keep cells
    new_min_pt = [3, 7]
    true_new_max_pt = [4, 9]

    new_discr = odl.uniform_discr_fromdiscr(discr, min_pt=new_min_pt)
    assert all_almost_equal(new_discr.min_pt, new_min_pt)
    assert all_almost_equal(new_discr.max_pt, true_new_max_pt)
    assert all_equal(new_discr.shape, discr.shape)
    assert all_almost_equal(new_discr.cell_sides, discr.cell_sides)

    # max_pt -> translate, keep cells
    new_max_pt = [3, 7]
    true_new_min_pt = [2, 5]

    new_discr = odl.uniform_discr_fromdiscr(discr, max_pt=new_max_pt)
    assert all_almost_equal(new_discr.min_pt, true_new_min_pt)
    assert all_almost_equal(new_discr.max_pt, new_max_pt)
    assert all_equal(new_discr.shape, discr.shape)
    assert all_almost_equal(new_discr.cell_sides, discr.cell_sides)

    # shape -> resize cells, keep corners
    new_shape = (5, 20)
    true_new_csides = [0.2, 0.1]
    new_discr = odl.uniform_discr_fromdiscr(discr, shape=new_shape)
    assert all_almost_equal(new_discr.min_pt, discr.min_pt)
    assert all_almost_equal(new_discr.max_pt, discr.max_pt)
    assert all_equal(new_discr.shape, new_shape)
    assert all_almost_equal(new_discr.cell_sides, true_new_csides)

    # cell_sides -> resize cells, keep corners
    new_csides = [0.5, 0.2]
    true_new_shape = (2, 10)
    new_discr = odl.uniform_discr_fromdiscr(discr, cell_sides=new_csides)
    assert all_almost_equal(new_discr.min_pt, discr.min_pt)
    assert all_almost_equal(new_discr.max_pt, discr.max_pt)
    assert all_equal(new_discr.shape, true_new_shape)
    assert all_almost_equal(new_discr.cell_sides, new_csides)


def test_uniform_discr_fromdiscr_two_attrs(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # Change 2 attributes -> resize and translate

    discr = odl.uniform_discr([0, -1], [1, 1], [10, 5], impl=impl, device=device)
    # csides = [0.1, 0.4]

    new_min_pt = [-2, 1]
    new_max_pt = [4, 2]
    true_new_csides = [0.6, 0.2]
    new_discr = odl.uniform_discr_fromdiscr(discr, min_pt=new_min_pt,
                                            max_pt=new_max_pt)
    assert all_almost_equal(new_discr.min_pt, new_min_pt)
    assert all_almost_equal(new_discr.max_pt, new_max_pt)
    assert all_equal(new_discr.shape, discr.shape)
    assert all_almost_equal(new_discr.cell_sides, true_new_csides)

    new_min_pt = [-2, 1]
    new_shape = (5, 20)
    true_new_max_pt = [-1.5, 9]
    new_discr = odl.uniform_discr_fromdiscr(discr, min_pt=new_min_pt,
                                            shape=new_shape)
    assert all_almost_equal(new_discr.min_pt, new_min_pt)
    assert all_almost_equal(new_discr.max_pt, true_new_max_pt)
    assert all_equal(new_discr.shape, new_shape)
    assert all_almost_equal(new_discr.cell_sides, discr.cell_sides)

    new_min_pt = [-2, 1]
    new_csides = [0.6, 0.2]
    true_new_max_pt = [4, 2]
    new_discr = odl.uniform_discr_fromdiscr(discr, min_pt=new_min_pt,
                                            cell_sides=new_csides)
    assert all_almost_equal(new_discr.min_pt, new_min_pt)
    assert all_almost_equal(new_discr.max_pt, true_new_max_pt)
    assert all_equal(new_discr.shape, discr.shape)
    assert all_almost_equal(new_discr.cell_sides, new_csides)

    new_max_pt = [4, 2]
    new_shape = (5, 20)
    true_new_min_pt = [3.5, -6]
    new_discr = odl.uniform_discr_fromdiscr(discr, max_pt=new_max_pt,
                                            shape=new_shape)
    assert all_almost_equal(new_discr.min_pt, true_new_min_pt)
    assert all_almost_equal(new_discr.max_pt, new_max_pt)
    assert all_equal(new_discr.shape, new_shape)
    assert all_almost_equal(new_discr.cell_sides, discr.cell_sides)

    new_max_pt = [4, 2]
    new_csides = [0.6, 0.2]
    true_new_min_pt = [-2, 1]
    new_discr = odl.uniform_discr_fromdiscr(discr, max_pt=new_max_pt,
                                            cell_sides=new_csides)
    assert all_almost_equal(new_discr.min_pt, true_new_min_pt)
    assert all_almost_equal(new_discr.max_pt, new_max_pt)
    assert all_equal(new_discr.shape, discr.shape)
    assert all_almost_equal(new_discr.cell_sides, new_csides)


def test_uniform_discr_fromdiscr_per_axis(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs

    discr = odl.uniform_discr([0, -1], [1, 1], [10, 5], impl=impl, device=device)
    # csides = [0.1, 0.4]

    new_min_pt = [-2, None]
    new_max_pt = [4, 2]
    new_shape = (None, 20)
    new_csides = [None, None]

    true_new_min_pt = [-2, -6]
    true_new_max_pt = [4, 2]
    true_new_shape = (10, 20)
    true_new_csides = [0.6, 0.4]

    new_discr = odl.uniform_discr_fromdiscr(
        discr, min_pt=new_min_pt, max_pt=new_max_pt,
        shape=new_shape, cell_sides=new_csides)

    assert all_almost_equal(new_discr.min_pt, true_new_min_pt)
    assert all_almost_equal(new_discr.max_pt, true_new_max_pt)
    assert all_equal(new_discr.shape, true_new_shape)
    assert all_almost_equal(new_discr.cell_sides, true_new_csides)

    new_min_pt = None
    new_max_pt = [None, 2]
    new_shape = (5, None)
    new_csides = [None, 0.2]

    true_new_min_pt = [0, 1]
    true_new_max_pt = [1, 2]
    true_new_shape = (5, 5)
    true_new_csides = [0.2, 0.2]

    new_discr = odl.uniform_discr_fromdiscr(
        discr, min_pt=new_min_pt, max_pt=new_max_pt,
        shape=new_shape, cell_sides=new_csides)

    assert all_almost_equal(new_discr.min_pt, true_new_min_pt)
    assert all_almost_equal(new_discr.max_pt, true_new_max_pt)
    assert all_equal(new_discr.shape, true_new_shape)
    assert all_almost_equal(new_discr.cell_sides, true_new_csides)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division
import pytest
import numpy as np

import odl
from odl.core.util.testutils import all_equal, all_almost_equal


# ---- RectPartition ---- #


def test_partition_init():
    vec1 = np.array([2, 4, 5, 7])
    vec2 = np.array([-4, -3, 0, 1, 4])
    min_pt = [2, -5]
    max_pt = [10, 4]

    # Simply test if code runs
    odl.RectPartition(odl.IntervalProd(min_pt, max_pt),
                      odl.RectGrid(vec1, vec2))
    odl.RectPartition(odl.IntervalProd(min_pt[0], max_pt[0]),
                      odl.RectGrid(vec1))

    # Degenerate dimensions should work, too
    vec2 = np.array([1.0])
    odl.RectPartition(odl.IntervalProd(min_pt, max_pt),
                      odl.RectGrid(vec1, vec2))


def test_partition_init_raise():
    # Check different error scenarios
    vec1 = np.array([2, 4, 5, 7])
    vec2 = np.array([-4, -3, 0, 1, 4])
    grid = odl.RectGrid(vec1, vec2)
    min_pt = [2, -5]
    max_pt = [10, 4]

    min_pt_toolarge = (2, -3.5)
    max_pt_toosmall = (7, 1)
    min_pt_badshape = (-1, 2, 0)
    max_pt_badshape = (2,)

    with pytest.raises(ValueError):
        odl.RectPartition(odl.IntervalProd(min_pt_toolarge, max_pt), grid)

    with pytest.raises(ValueError):
        odl.RectPartition(odl.IntervalProd(min_pt, max_pt_toosmall), grid)

    with pytest.raises(ValueError):
        odl.RectPartition(odl.IntervalProd(min_pt_badshape, max_pt_badshape),
                          grid)

    with pytest.raises(TypeError):
        odl.RectPartition(None, grid)

    with pytest.raises(TypeError):
        odl.RectPartition(odl.IntervalProd(min_pt_toolarge, max_pt), None)


def _test_eq(x, y):
    """Test equality of x and y."""
    assert x == y
    assert not x != y
    assert hash(x) == hash(y)


def _test_neq(x, y):
    """Test non-equality of x and y."""
    assert x != y
    assert not x == y
    assert hash(x) != hash(y)


def test_equals():
    """Test partition equality checks and hash."""
    vec1 = np.array([2, 3, 4, 5])
    vec2 = np.array([-4, -2, 0, 2, 4])

    part1 = odl.RectPartition(odl.IntervalProd(1, 6),
                              odl.RectGrid(vec1))
    part2 = odl.RectPartition(odl.IntervalProd([1, -5], [6, 5]),
                              odl.RectGrid(vec1, vec2))
    part2_again = odl.RectPartition(odl.IntervalProd([1, -5], [6, 5]),
                                    odl.RectGrid(vec1, vec2))
    part2_rev = odl.RectPartition(odl.IntervalProd([-5, 1], [5, 6]),
                                  odl.RectGrid(vec2, vec1))

    _test_eq(part1, part1)
    _test_eq(part2, part2)
    _test_eq(part2, part2_again)

    _test_neq(part1, part2)
    _test_neq(part2, part2_rev)
    assert part2 != (vec1, vec2)


def test_partition_set():
    vec1 = np.array([2, 4, 5, 7])
    vec2 = np.array([-4, -3, 0, 1, 4])
    grid = odl.RectGrid(vec1, vec2)

    min_pt = [1, -4]
    max_pt = [10, 5]
    intv = odl.IntervalProd(min_pt, max_pt)

    part = odl.RectPartition(intv, grid)
    assert part.set == odl.IntervalProd(min_pt, max_pt)
    assert all_equal(part.min_pt, min_pt)
    assert all_equal(part.min(), min_pt)
    assert all_equal(part.max_pt, max_pt)
    assert all_equal(part.max(), max_pt)


def test_partition_cell_boundary_vecs():
    vec1 = np.array([2, 4, 5, 7])
    vec2 = np.array([-4, -3, 0, 1, 4])
    grid = odl.RectGrid(vec1, vec2)

    midpts1 = [3, 4.5, 6]
    midpts2 = [-3.5, -1.5, 0.5, 2.5]

    min_pt = [2, -6]
    max_pt = [10, 4]
    intv = odl.IntervalProd(min_pt, max_pt)

    true_bvec1 = [2] + midpts1 + [10]
    true_bvec2 = [-6] + midpts2 + [4]

    part = odl.RectPartition(intv, grid)
    assert all_equal(part.cell_boundary_vecs, (true_bvec1, true_bvec2))


def test_partition_cell_sizes_vecs():
    vec1 = np.array([2, 4, 5, 7])
    vec2 = np.array([-4, -3, 0, 1, 4])
    grid = odl.RectGrid(vec1, vec2)

    midpts1 = [3, 4.5, 6]
    midpts2 = [-3.5, -1.5, 0.5, 2.5]

    min_pt = [2, -6]
    max_pt = [10, 4]
    intv = odl.IntervalProd(min_pt, max_pt)

    bvec1 = np.array([2] + midpts1 + [10])
    bvec2 = np.array([-6] + midpts2 + [4])
    true_csizes1 = bvec1[1:] - bvec1[:-1]
    true_csizes2 = bvec2[1:] - bvec2[:-1]

    part = odl.RectPartition(intv, grid)
    assert all_equal(part.cell_sizes_vecs, (true_csizes1, true_csizes2))


def test_partition_cell_sides():
    grid = odl.uniform_grid([0, 1], [2, 4], (5, 3))
    intv = odl.IntervalProd([0, 1], [2, 4])
    part = odl.RectPartition(intv, grid)
    true_sides = [0.5, 1.5]
    assert all_equal(part.cell_sides, true_sides)


def test_partition_cell_volume():
    grid = odl.uniform_grid([0, 1], [2, 4], (5, 3))
    intv = odl.IntervalProd([0, 1], [2, 4])
    part = odl.RectPartition(intv, grid)
    true_volume = 0.5 * 1.5
    assert part.cell_volume == true_volume


def test_partition_insert():
    vec11 = [2, 4, 5, 7]
    vec12 = [-4, -3, 0, 1, 4]
    min_pt1 = [1, -4]
    max_pt1 = [7, 5]
    grid1 = odl.RectGrid(vec11, vec12)
    intv1 = odl.IntervalProd(min_pt1, max_pt1)
    part1 = odl.RectPartition(intv1, grid1)

    vec21 = [-2, 0, 3]
    vec22 = [0]
    min_pt2 = [-2, -2]
    max_pt2 = [4, 0]
    grid2 = odl.RectGrid(vec21, vec22)
    intv2 = odl.IntervalProd(min_pt2, max_pt2)
    part2 = odl.RectPartition(intv2, grid2)

    part = part1.insert(0, part2)
    assert all_equal(part.min_pt, [-2, -2, 1, -4])
    assert all_equal(part.max_pt, [4, 0, 7, 5])
    assert all_equal(part.grid.min_pt, [-2, 0, 2, -4])
    assert all_equal(part.grid.max_pt, [3, 0, 7, 4])

    part = part1.insert(1, part2)
    assert all_equal(part.min_pt, [1, -2, -2, -4])
    assert all_equal(part.max_pt, [7, 4, 0, 5])
    assert all_equal(part.grid.min_pt, [2, -2, 0, -4])
    assert all_equal(part.grid.max_pt, [7, 3, 0, 4])


def test_partition_getitem():
    vec1 = [2, 4, 5, 7]
    vec2 = [-4, -3, 0, 1, 4]
    vec3 = [-2, 0, 3]
    vec4 = [0]
    vecs = [vec1, vec2, vec3, vec4]
    min_pt = [1, -4, -2, -2]
    max_pt = [7, 5, 4, 0]
    grid = odl.RectGrid(*vecs)
    intv = odl.IntervalProd(min_pt, max_pt)
    part = odl.RectPartition(intv, grid)

    # Test a couple of slices
    slc = (1, -2, 2, 0)
    slc_vecs = [v[i] for i, v in zip(slc, vecs)]
    slc_part = part[slc]
    assert slc_part.grid == odl.RectGrid(*slc_vecs)
    slc_min_pt = [3, 0.5, 1.5, -2]
    slc_max_pt = [4.5, 2.5, 4, 0]
    assert slc_part.set == odl.IntervalProd(slc_min_pt, slc_max_pt)

    slc = (slice(None), slice(None, None, 2), slice(None, 2), 0)
    slc_vecs = [v[i] for i, v in zip(slc, vecs)]
    slc_part = part[slc]
    assert slc_part.grid == odl.RectGrid(*slc_vecs)
    slc_min_pt = [1, -4, -2, -2]
    slc_max_pt = [7, 5, 1.5, 0]
    assert slc_part.set == odl.IntervalProd(slc_min_pt, slc_max_pt)

    # Fewer indices
    assert part[1] == part[1, :, :, :] == part[1, ...]
    assert part[1, 2:] == part[1, 2:, :, :] == part[1, 2:, ...]
    assert part[1, 2:, ::2] == part[1, 2:, ::2, :] == part[1, 2:, ::2, ...]

    # Index list using indices 0 and 2
    lst_min_pt = [1, -4, -2, -2]
    lst_max_pt = [6, 5, 4, 0]
    lst_intv = odl.IntervalProd(lst_min_pt, lst_max_pt)
    lst_vec1 = [2, 5]
    lst_grid = odl.RectGrid(lst_vec1, vec2, vec3, vec4)
    lst_part = odl.RectPartition(lst_intv, lst_grid)
    assert part[[0, 2]] == lst_part


def test_partition_byaxis():
    """Test indexing a partition along the axes."""
    part = odl.uniform_partition([-1, -2, -3, -4], [1, 2, 3, 4], (1, 2, 4, 5))

    assert part.byaxis[0] == odl.uniform_partition(-1, 1, 1)
    assert part.byaxis[1:3] == odl.uniform_partition([-2, -3], [2, 3], (2, 4))
    assert (part.byaxis[[1, 3, 1]] ==
            odl.uniform_partition([-2, -4, -2], [2, 4, 2], (2, 5, 2)))
    assert part.byaxis[[1]] == part.byaxis[1]
    assert part.byaxis[...] == part
    assert part.byaxis[:] == part


def test_empty_partition():
    """Check if empty partitions behave as expected and all methods work."""
    part = odl.RectPartition(odl.IntervalProd([], []),
                             odl.uniform_grid([], [], ()))

    assert part.cell_boundary_vecs == ()
    assert part.nodes_on_bdry is True
    assert part.nodes_on_bdry_byaxis == ()
    assert part.has_isotropic_cells
    assert part.boundary_cell_fractions == ()
    assert part.cell_sizes_vecs == ()
    assert np.array_equal(part.cell_sides, [])
    assert part.cell_volume == 0

    same = odl.RectPartition(odl.IntervalProd([], []),
                             odl.uniform_grid([], [], ()))
    assert part == same
    assert hash(part) == hash(same)
    other = odl.uniform_partition(0, 1, 4)
    assert part != other

    assert part[[]] == part
    assert part.insert(0, other) == other
    assert other.insert(0, part) == other
    assert other.insert(1, part) == other
    assert part.squeeze() == part
    assert part.index([]) == ()
    part.byaxis
    assert part == odl.uniform_partition([], [], ())
    repr(part)


# ---- Functions ---- #


def test_uniform_partition_fromintv():
    intvp = odl.IntervalProd([0, 0], [1, 2])
    shape = (4, 10)

    # All nodes at the boundary
    part = odl.uniform_partition_fromintv(intvp, shape, nodes_on_bdry=True)
    assert all_equal(part.min_pt, intvp.min_pt)
    assert all_equal(part.max_pt, intvp.max_pt)
    assert all_equal(part.grid.min_pt, intvp.min_pt)
    assert all_equal(part.grid.max_pt, intvp.max_pt)
    for cs in part.cell_sizes_vecs:
        # Check that all cell sizes are equal (except first and last which
        # are halved)
        assert np.allclose(np.diff(cs[1:-1]), 0)
        assert all_almost_equal(cs[0], cs[1] / 2)
        assert all_almost_equal(cs[-1], cs[-2] / 2)

    # All nodes not the boundary
    part = odl.uniform_partition_fromintv(intvp, shape, nodes_on_bdry=False)
    assert all_equal(part.min_pt, intvp.min_pt)
    assert all_equal(part.max_pt, intvp.max_pt)
    for cs in part.cell_sizes_vecs:
        # Check that all cell sizes are equal
        assert np.allclose(np.diff(cs), 0)

    # Only left nodes at the boundary
    part = odl.uniform_partition_fromintv(intvp, shape,
                                          nodes_on_bdry=[[True, False]] * 2)
    assert all_equal(part.min_pt, intvp.min_pt)
    assert all_equal(part.max_pt, intvp.max_pt)
    assert all_equal(part.grid.min_pt, intvp.min_pt)
    for cs in part.cell_sizes_vecs:
        # Check that all cell sizes are equal (except first)
        assert np.allclose(np.diff(cs[1:]), 0)
        assert all_almost_equal(cs[0], cs[1] / 2)

    # Only right nodes at the boundary
    part = odl.uniform_partition_fromintv(intvp, shape,
                                          nodes_on_bdry=[[False, True]] * 2)
    assert all_equal(part.min_pt, intvp.min_pt)
    assert all_equal(part.max_pt, intvp.max_pt)
    assert all_equal(part.grid.max_pt, intvp.max_pt)
    for cs in part.cell_sizes_vecs:
        # Check that all cell sizes are equal (except last)
        assert np.allclose(np.diff(cs[:-1]), 0)
        assert all_almost_equal(cs[-1], cs[-2] / 2)


def test_uniform_partition_fromgrid():
    vec1 = np.array([2, 4, 5, 7])
    vec2 = np.array([-4, -3, 0, 1, 4])
    min_pt = [0, -4]
    max_pt = [7, 8]
    min_pt_calc = [2 - (4 - 2) / 2, -4 - (-3 + 4) / 2]
    max_pt_calc = [7 + (7 - 5) / 2, 4 + (4 - 1) / 2]

    # Default case
    grid = odl.RectGrid(vec1, vec2)
    part = odl.uniform_partition_fromgrid(grid)
    assert part.set == odl.IntervalProd(min_pt_calc, max_pt_calc)

    # Explicit min_pt / max_pt, full vectors
    part = odl.uniform_partition_fromgrid(grid, min_pt=min_pt)
    assert part.set == odl.IntervalProd(min_pt, max_pt_calc)
    part = odl.uniform_partition_fromgrid(grid, max_pt=max_pt)
    assert part.set == odl.IntervalProd(min_pt_calc, max_pt)

    # min_pt / max_pt as dictionaries
    min_pt_dict = {0: 0.5}
    max_pt_dict = {-1: 8}
    part = odl.uniform_partition_fromgrid(
        grid, min_pt=min_pt_dict, max_pt=max_pt_dict)
    true_min_pt = [0.5, min_pt_calc[1]]
    true_max_pt = [max_pt_calc[0], 8]
    assert part.set == odl.IntervalProd(true_min_pt, true_max_pt)

    # Degenerate dimension, needs both explicit min_pt and max_pt
    grid = odl.RectGrid(vec1, [1.0])
    with pytest.raises(ValueError):
        odl.uniform_partition_fromgrid(grid)
    with pytest.raises(ValueError):
        odl.uniform_partition_fromgrid(grid, min_pt=min_pt)
    with pytest.raises(ValueError):
        odl.uniform_partition_fromgrid(grid, max_pt=max_pt)


def test_uniform_partition():

    min_pt = [0, 0]
    max_pt = [1, 2]
    shape = (4, 10)
    csides = [0.25, 0.2]

    # Test standard case
    part = odl.uniform_partition(min_pt, max_pt, shape, nodes_on_bdry=True)

    assert all_equal(part.min_pt, min_pt)
    assert all_equal(part.max_pt, max_pt)
    assert all_equal(part.grid.min_pt, min_pt)
    assert all_equal(part.grid.max_pt, max_pt)
    for cs in part.cell_sizes_vecs:
        # Check that all cell sizes are equal (except first and last which
        # are halved)
        assert np.allclose(np.diff(cs[1:-1]), 0)
        assert all_almost_equal(cs[0], cs[1] / 2)
        assert all_almost_equal(cs[-1], cs[-2] / 2)

    assert part[1:, 2:5].is_uniform
    assert part[1:, ::3].is_uniform

    # Test combinations of parameters
    true_part = odl.uniform_partition(min_pt, max_pt, shape,
                                      nodes_on_bdry=False)
    part = odl.uniform_partition(min_pt=min_pt, max_pt=max_pt, shape=shape,
                                 cell_sides=None)
    assert part == true_part
    part = odl.uniform_partition(min_pt=min_pt, max_pt=max_pt, shape=None,
                                 cell_sides=csides)
    assert part == true_part
    part = odl.uniform_partition(min_pt=min_pt, max_pt=None,
                                 shape=shape, cell_sides=csides)
    assert part == true_part
    part = odl.uniform_partition(min_pt=None, max_pt=max_pt, shape=shape,
                                 cell_sides=csides)
    assert part == true_part
    part = odl.uniform_partition(min_pt=min_pt, max_pt=max_pt, shape=shape,
                                 cell_sides=csides)
    assert part == true_part

    # Test parameters per axis
    part = odl.uniform_partition(
        min_pt=[0, None], max_pt=[None, 2], shape=shape, cell_sides=csides)
    assert part == true_part
    part = odl.uniform_partition(
        min_pt=min_pt, max_pt=[None, 2], shape=(4, None), cell_sides=csides)
    assert part == true_part
    part = odl.uniform_partition(
        min_pt=min_pt, max_pt=max_pt, shape=(None, 4), cell_sides=[0.25, None])

    # Test robustness against numerical error
    part = odl.uniform_partition(
        min_pt=min_pt, max_pt=[None, np.sqrt(2) ** 2], shape=shape,
        cell_sides=[0.25, np.log(np.exp(0.2))])
    assert part.approx_equals(true_part, atol=1e-8)

    # Test nodes_on_bdry
    # Here we compute stuff, so we can only expect approximate equality
    csides = [1 / 3., 2 / 9.5]
    true_part = odl.uniform_partition(min_pt, max_pt, shape,
                                      nodes_on_bdry=(True, (False, True)))
    part = odl.uniform_partition(
        min_pt=[0, None], max_pt=[None, 2], shape=shape,
        cell_sides=csides, nodes_on_bdry=(True, (False, True)))
    assert part.approx_equals(true_part, atol=1e-8)
    part = odl.uniform_partition(
        min_pt=min_pt, max_pt=[None, 2], shape=(4, None), cell_sides=csides,
        nodes_on_bdry=(True, (False, True)))
    assert part.approx_equals(true_part, atol=1e-8)
    part = odl.uniform_partition(
        min_pt=min_pt, max_pt=max_pt, shape=(None, 10),
        cell_sides=[1 / 3., None],
        nodes_on_bdry=(True, (False, True)))
    assert part.approx_equals(true_part, atol=1e-8)

    # Test error scenarios

    # Not enough parameters (total / per axis)
    with pytest.raises(ValueError):
        odl.uniform_partition()

    with pytest.raises(ValueError):
        odl.uniform_partition(min_pt, max_pt)

    with pytest.raises(ValueError):
        part = odl.uniform_partition(
            min_pt=[0, None], max_pt=[1, None], shape=shape,
            cell_sides=csides)

    with pytest.raises(ValueError):
        part = odl.uniform_partition(
            min_pt=min_pt, max_pt=[1, None], shape=(4, None),
            cell_sides=csides)

    # Parameters with inconsistent sizes
    with pytest.raises(ValueError):
        part = odl.uniform_partition(
            min_pt=min_pt, max_pt=[1, None, None], shape=shape)

    # Too large rounding error in computing shape
    with pytest.raises(ValueError):
        part = odl.uniform_partition(
            min_pt=min_pt, max_pt=max_pt, cell_sides=[0.25, 0.2001])

    # Inconsistent values
    with pytest.raises(ValueError):
        part = odl.uniform_partition(
            min_pt=min_pt, max_pt=max_pt, shape=shape,
            cell_sides=[0.25, 0.2001])


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



import pytest

import odl
from odl.core.util.pytest_config import IMPL_DEVICE_PAIRS
from odl.core.util.testutils import all_almost_equal

try:
    import torch
except ImportError:
    pass

skip_if_no_pytorch = pytest.mark.skipif(
        "'pytorch' not in odl.core.space.entry_points.TENSOR_SPACE_IMPLS",
        reason='PYTORCH not available',
    )

IMPLS = [pytest.param(value, marks=skip_if_no_pytorch) for value in IMPL_DEVICE_PAIRS]

DEFAULT_SHAPE = (4,4)

@pytest.fixture(scope='module', params=IMPLS)
def tspace(request, odl_floating_dtype):
    impl, device = request.param
    return odl.tensor_space(
        shape=DEFAULT_SHAPE, 
        dtype=odl_floating_dtype, 
        impl=impl, 
        device=device
    )

@pytest.fixture(scope='module')
def numpy_tspace(odl_floating_dtype):
    return odl.tensor_space(
        shape=DEFAULT_SHAPE, 
        dtype=odl_floating_dtype, 
        impl='numpy', 
        device='cpu'
    )

@pytest.fixture(scope='module')
def pytorch_tspace_cpu(odl_floating_dtype):
    return odl.tensor_space(
        shape=DEFAULT_SHAPE, 
        dtype=odl_floating_dtype, 
        impl='pytorch', 
        device='cpu'
    )

@pytest.fixture(scope='module')
def pytorch_tspace_gpu(odl_floating_dtype):
    if torch.cuda.device_count() == 0:
        pytest.skip(reason="No Cuda-capable GPU available")

    return odl.tensor_space(
        shape=DEFAULT_SHAPE, 
        dtype=odl_floating_dtype, 
        impl='pytorch', 
        device='cuda:0'
    )

def test_same_backend_same_device(tspace, odl_arithmetic_op):
    """Test that operations between two elements on separate spaces with the same backend are possible"""
    x = next(tspace.examples)[1]
    y = next(tspace.examples)[1]
    op = odl_arithmetic_op
    z_arr = op(x.data, y.data)
    z = op(x, y)
    assert all_almost_equal([x, y, z], [x.data, y.data, z_arr])

@skip_if_no_pytorch
def test_different_backends(
        numpy_tspace, pytorch_tspace_cpu, pytorch_tspace_gpu,
        odl_arithmetic_op
        ):
    """Test that operations between two elements on separate spaces with different device or impl are not possible"""
    x_np = next(numpy_tspace.examples)[1]
    x_pt_cpu = next(pytorch_tspace_cpu.examples)[1]
    x_pt_gpu = next(pytorch_tspace_gpu.examples)[1]
    op = odl_arithmetic_op

    # Same device, different backend
    with pytest.raises(AssertionError):
        res = op(x_np, x_pt_cpu) 
    
    with pytest.raises(TypeError):
        res = op(x_np, x_pt_cpu.data) 

    with pytest.raises(TypeError):
        res = op(x_np.data, x_pt_cpu) 

    # Same backend, different device
    with pytest.raises(AssertionError):
        res = op(x_pt_gpu, x_pt_cpu) 
    
    with pytest.raises(TypeError):
        res = op(x_pt_gpu.data, x_pt_cpu) 

    with pytest.raises(TypeError):
        res = op(x_pt_gpu, x_pt_cpu.data) 

    # Different device, different backend
    with pytest.raises(AssertionError):
        res = op(x_np, x_pt_gpu) 

    with pytest.raises(TypeError):
        res = op(x_np, x_pt_gpu.data) 

    with pytest.raises(TypeError):
        res = op(x_np.data, x_pt_gpu) 


    


# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Utility functions for Fourier transforms on regularly sampled data."""

from __future__ import absolute_import, division, print_function

import numpy as np

from odl.core.util.npy_compat import AVOID_UNNECESSARY_COPY

from odl.core.discr import (
    DiscretizedSpace, uniform_discr_frompartition, uniform_grid,
    uniform_partition_fromgrid)
from odl.core.set import RealNumbers
from odl.core.util import (
    complex_dtype, conj_exponent, dtype_repr, fast_1d_tensor_mult,
    is_complex_dtype, is_numeric_dtype, is_real_dtype,
    is_real_floating_dtype, is_string, normalized_axes_tuple,
    normalized_scalar_param_list)
from odl.core.array_API_support import get_array_and_backend, ArrayBackend

from odl.core.util.dtype_utils import _universal_dtype_identifier

__all__ = ('reciprocal_grid', 'realspace_grid',
           'reciprocal_space',
           'dft_preprocess_data', 'dft_postprocess_data')


def reciprocal_grid(grid, shift=True, axes=None, halfcomplex=False):
    """Return the reciprocal of the given regular grid.

    This function calculates the reciprocal (Fourier/frequency space)
    grid for a given regular grid defined by the nodes::

        x[k] = x[0] + k * s,

    where ``k = (k[0], ..., k[d-1])`` is a ``d``-dimensional index in
    the range ``0 <= k < N`` (component-wise). The multi-index
    ``N`` is the shape of the input grid.
    This grid's reciprocal is then given by the nodes::

        xi[j] = xi[0] + j * sigma,

    with the reciprocal grid stride ``sigma = 2*pi / (s * N)``.
    The minimum frequency ``xi[0]`` can in principle be chosen
    freely, but usually it is chosen in a such a way that the reciprocal
    grid is centered around zero. For this, there are two possibilities:

    1. Make the grid point-symmetric around 0.

    2. Make the grid "almost" point-symmetric around zero by shifting
       it to the left by half a reciprocal stride.

    In the first case, the minimum frequency (per axis) is given as::

        xi_1[0] = -pi/s + pi/(s*n) = -pi/s + sigma/2.

    For the second case, it is::

        xi_1[0] = -pi / s.

    Note that the zero frequency is contained in case 1 for an odd
    number of points, while for an even size, the second option
    guarantees that 0 is contained.

    If a real-to-complex (half-complex) transform is to be computed,
    the reciprocal grid has the shape ``M[i] = floor(N[i]/2) + 1``
    in the last transform axis ``i``.

    Parameters
    ----------
    grid : uniform `RectGrid`
        Original sampling grid,.
    shift : bool or sequence of bools, optional
        If ``True``, the grid is shifted by half a stride in the negative
        direction. With a sequence, this option is applied separately on
        each axis.
    axes : int or sequence of ints, optional
        Dimensions in which to calculate the reciprocal. The sequence
        must have the same length as ``shift`` if the latter is given
        as a sequence. ``None`` means all axes in ``grid``.
    halfcomplex : bool, optional
        If ``True``, return the half of the grid with last coordinate
        less than zero. This is related to the fact that for real-valued
        functions, the other half is the mirrored complex conjugate of
        the given half and therefore needs not be stored.

    Returns
    -------
    reciprocal_grid : uniform `RectGrid`
        The reciprocal grid.
    """
    if axes is None:
        axes = list(range(grid.ndim))
    else:
        try:
            axes = [int(axes)]
        except TypeError:
            axes = list(axes)

    # List indicating shift or not per "active" axis, same length as axes
    shift_list = normalized_scalar_param_list(shift, length=len(axes),
                                              param_conv=bool)

    # Full-length vectors
    stride = grid.stride.copy()
    stride[stride == 0] = 1
    shape = np.array(grid.shape)
    rmin = grid.min_pt.copy()
    rmax = grid.max_pt.copy()
    rshape = list(shape)

    # Shifted axes (full length to avoid ugly double indexing)
    shifted = np.zeros(grid.ndim, dtype=bool)
    shifted[axes] = shift_list
    rmin[shifted] = -np.pi / stride[shifted]
    # Length min->max increases by double the shift, so we
    # have to compensate by a full stride
    rmax[shifted] = (-rmin[shifted] -
                     2 * np.pi / (stride[shifted] * shape[shifted]))

    # Non-shifted axes
    not_shifted = np.zeros(grid.ndim, dtype=bool)
    not_shifted[axes] = np.logical_not(shift_list)
    rmin[not_shifted] = ((-1.0 + 1.0 / shape[not_shifted]) *
                         np.pi / stride[not_shifted])
    rmax[not_shifted] = -rmin[not_shifted]

    # Change last axis shape and max if halfcomplex
    if halfcomplex:
        rshape[axes[-1]] = shape[axes[-1]] // 2 + 1

        # - Odd and shifted: - stride / 2
        # - Even and not shifted: + stride / 2
        # - Otherwise: 0
        last_odd = shape[axes[-1]] % 2 == 1
        last_shifted = shift_list[-1]
        half_rstride = np.pi / (shape[axes[-1]] * stride[axes[-1]])

        if last_odd and last_shifted:
            rmax[axes[-1]] = -half_rstride
        elif not last_odd and not last_shifted:
            rmax[axes[-1]] = half_rstride
        else:
            rmax[axes[-1]] = 0

    return uniform_grid(rmin, rmax, rshape)


def realspace_grid(recip_grid, x0, axes=None, halfcomplex=False,
                   halfcx_parity='even'):
    """Return the real space grid from the given reciprocal grid.

    Given a reciprocal grid::

        xi[j] = xi[0] + j * sigma,

    with a multi-index ``j = (j[0], ..., j[d-1])`` in the range
    ``0 <= j < M``, this function calculates the original grid::

        x[k] = x[0] + k * s

    by using a provided ``x[0]`` and calculating the stride ``s``.

    If the reciprocal grid is interpreted as coming from a usual
    complex-to-complex FFT, it is ``N == M``, and the stride is::

        s = 2*pi / (sigma * N)

    For a reciprocal grid from a real-to-complex (half-complex) FFT,
    it is ``M[i] = floor(N[i]/2) + 1`` in the last transform axis ``i``.
    To resolve the ambiguity regarding the parity of ``N[i]``, the
    it must be specified if the output shape should be even or odd,
    resulting in::

        odd : N[i] = 2 * M[i] - 1
        even: N[i] = 2 * M[i] - 2

    The output stride is calculated with this ``N`` as above in this
    case.

    Parameters
    ----------
    recip_grid : uniform `RectGrid`
        Sampling grid in reciprocal space.
    x0 : `array-like`
        Desired minimum point of the real space grid.
    axes : int or sequence of ints, optional
        Dimensions in which to calculate the real space grid. The sequence
        must have the same length as ``shift`` if the latter is given
        as a sequence. ``None`` means "all axes".
    halfcomplex : bool, optional
        If ``True``, interpret the given grid as the reciprocal as used
        in a half-complex FFT (see above). Otherwise, the grid is
        regarded as being used in a complex-to-complex transform.
    halfcx_parity : {'even', 'odd'}
        Use this parity for the shape of the returned grid in the
        last axis of ``axes`` in the case ``halfcomplex=True``

    Returns
    -------
    irecip : uniform `RectGrid`
        The inverse reciprocal grid.
    """
    if axes is None:
        axes = list(range(recip_grid.ndim))
    else:
        try:
            axes = [int(axes)]
        except TypeError:
            axes = list(axes)

    rstride = recip_grid.stride
    rshape = recip_grid.shape

    # Calculate shape of the output grid by adjusting in axes[-1]
    irshape = list(rshape)
    if halfcomplex:
        if str(halfcx_parity).lower() == 'even':
            irshape[axes[-1]] = 2 * rshape[axes[-1]] - 2
        elif str(halfcx_parity).lower() == 'odd':
            irshape[axes[-1]] = 2 * rshape[axes[-1]] - 1
        else:
            raise ValueError("`halfcomplex` parity '{}' not understood"
                             "".format(halfcx_parity))

    irmin = np.asarray(x0)
    irshape = np.asarray(irshape)
    irstride = np.copy(rstride)
    irstride[axes] = 2 * np.pi / (irshape[axes] * rstride[axes])
    irmax = irmin + (irshape - 1) * irstride

    return uniform_grid(irmin, irmax, irshape)


def dft_preprocess_data(arr, shift=True, axes=None, sign='-', out=None):
    """Pre-process the real-space data before DFT.

    This function multiplies the given data with the separable
    function::

        p(x) = exp(+- 1j * dot(x - x[0], xi[0]))

    where ``x[0]`` and ``xi[0]`` are the minimum coodinates of
    the real-space and reciprocal grids, respectively. The sign of
    the exponent depends on the choice of ``sign``. In discretized
    form, this function becomes an array::

        p[k] = exp(+- 1j * k * s * xi[0])

    If the reciprocal grid is not shifted, i.e. symmetric around 0,
    it is ``xi[0] =  pi/s * (-1 + 1/N)``, hence::

        p[k] = exp(-+ 1j * pi * k * (1 - 1/N))

    For a shifted grid, we have :math:``xi[0] =  -pi/s``, thus the
    array is given by::

        p[k] = (-1)**k

    Parameters
    ----------
    arr : `array-like`
        Array to be pre-processed. If its data type is a real
        non-floating type, it is converted to 'float64'.
    shift : bool or or sequence of bools, optional
        If ``True``, the grid is shifted by half a stride in the negative
        direction. With a sequence, this option is applied separately on
        each axis.
    axes : int or sequence of ints, optional
        Dimensions in which to calculate the reciprocal. The sequence
        must have the same length as ``shift`` if the latter is given
        as a sequence.
        Default: all axes.
    sign : {'-', '+'}, optional
        Sign of the complex exponent.
    out : `numpy.ndarray`, optional
        Array in which the result is stored. If ``out is arr``,
        an in-place modification is performed. For real data type,
        this is only possible for ``shift=True`` since the factors are
        complex otherwise.

    Returns
    -------
    out : `numpy.ndarray`
        Result of the pre-processing. If ``out`` was given, the returned
        object is a reference to it.

    Notes
    -----
    If ``out`` is not specified, the data type of the returned array
    is the same as that of ``arr`` except when ``arr`` has real data
    type and ``shift`` is not ``True``. In this case, the return type
    is the complex counterpart of ``arr.dtype``.
    """
    arr, backend = get_array_and_backend(arr)
    backend : ArrayBackend
    dtype = backend.get_dtype_identifier(array=arr)
    if not is_numeric_dtype(arr.dtype):
        raise ValueError('array has non-numeric data type {}'
                         ''.format(dtype_repr(arr.dtype)))
    elif is_real_dtype(arr.dtype) and not is_real_floating_dtype(arr.dtype):
        arr = arr.astype('float64')

    if axes is None:
        axes = list(range(arr.ndim))
    else:
        try:
            axes = [int(axes)]
        except TypeError:
            axes = list(axes)

    shape = arr.shape
    shift_list = normalized_scalar_param_list(shift, length=len(axes),
                                              param_conv=bool)

    # Make a copy of arr with correct data type if necessary, or copy values.
    if out is None:
        if all(shift_list):
            dtype = backend.available_dtypes[dtype]
        else:
            dtype = backend.available_dtypes[complex_dtype(dtype)]

        out = backend.array_constructor(
            arr, dtype=dtype, copy=True, device=arr.device)
    else:
        out[:] = arr

    if is_real_dtype(out.dtype) and not shift:
        raise ValueError('cannot pre-process real input in-place without '
                         'shift')

    if sign == '-':
        imag = -1j
    elif sign == '+':
        imag = 1j
    else:
        raise ValueError("`sign` '{}' not understood".format(sign))
    
    out_dtype = _universal_dtype_identifier(out.dtype)

    def _onedim_arr(length, shift):
        if shift:
            # (-1)^indices
            factor = np.ones(length, dtype=out_dtype)
            factor[1::2] = -1
        else:
            factor = np.arange(length, dtype=out_dtype)
            factor *= -imag * np.pi * (1 - 1.0 / length)
            np.exp(factor, out=factor)
        return factor.astype(out_dtype, copy=AVOID_UNNECESSARY_COPY)

    onedim_arrs = []
    for axis, shift in zip(axes, shift_list):
        length = shape[axis]
        onedim_arrs.append(_onedim_arr(length, shift))

    fast_1d_tensor_mult(out, onedim_arrs, axes=axes, out=out)
    return out


def _interp_kernel_ft(norm_freqs, interp):
    """Scaled FT of a one-dimensional interpolation kernel.

    For normalized frequencies ``-1/2 <= xi <= 1/2``, this
    function returns::

        sinc(pi * xi)**k / sqrt(2 * pi)

    where ``k=1`` for 'nearest' and ``k=2`` for 'linear' interpolation.

    Parameters
    ----------
    norm_freqs : `numpy.ndarray`
        Normalized frequencies between -1/2 and 1/2
    interp : {'nearest', 'linear'}
        Type of interpolation kernel

    Returns
    -------
    ker_ft : `numpy.ndarray`
        Values of the kernel FT at the given frequencies
    """
    # Numpy's sinc(x) is equal to the 'math' sinc(pi * x)
    ker_ft = np.sinc(norm_freqs)
    interp_ = str(interp).lower()
    if interp_ == 'nearest':
        pass
    elif interp_ == 'linear':
        ker_ft *= ker_ft
    else:
        raise ValueError("`interp` '{}' not understood".format(interp))

    ker_ft /= np.sqrt(2 * np.pi)
    return ker_ft


def dft_postprocess_data(arr, real_grid, recip_grid, shift, axes,
                         interp, sign='-', op='multiply', out=None):
    """Post-process the Fourier-space data after DFT.

    This function multiplies the given data with the separable
    function::

        q(xi) = exp(+- 1j * dot(x[0], xi)) * s * phi_hat(xi_bar)

    where ``x[0]`` and ``s`` are the minimum point and the stride of
    the real-space grid, respectively, and ``phi_hat(xi_bar)`` is the FT
    of the interpolation kernel. The sign of the exponent depends on the
    choice of ``sign``. Note that for ``op='divide'`` the
    multiplication with ``s * phi_hat(xi_bar)`` is replaced by a
    division with the same array.

    In discretized form on the reciprocal grid, the exponential part
    of this function becomes an array::

        q[k] = exp(+- 1j * dot(x[0], xi[k]))

    and the arguments ``xi_bar`` to the interpolation kernel
    are the normalized frequencies::

        for 'shift=True'  : xi_bar[k] = -pi + pi * (2*k) / N
        for 'shift=False' : xi_bar[k] = -pi + pi * (2*k+1) / N

    See [Pre+2007], Section 13.9 "Computing Fourier Integrals Using
    the FFT" for a similar approach.

    Parameters
    ----------
    arr : `array-like`
        Array to be pre-processed. An array with real data type is
        converted to its complex counterpart.
    real_grid : uniform `RectGrid`
        Real space grid in the transform.
    recip_grid : uniform `RectGrid`
        Reciprocal grid in the transform
    shift : bool or sequence of bools
        If ``True``, the grid is shifted by half a stride in the negative
        direction in the corresponding axes. The sequence must have the
        same length as ``axes``.
    axes : int or sequence of ints
        Dimensions along which to take the transform. The sequence must
        have the same length as ``shifts``.
    interp : string or sequence of strings
        Interpolation scheme used in the real-space.
    sign : {'-', '+'}, optional
        Sign of the complex exponent.
    op : {'multiply', 'divide'}, optional
        Operation to perform with the stride times the interpolation
        kernel FT
    out : `numpy.ndarray`, optional
        Array in which the result is stored. If ``out is arr``, an
        in-place modification is performed.

    Returns
    -------
    out : `numpy.ndarray`
        Result of the post-processing. If ``out`` was given, the returned
        object is a reference to it.

    References
    ----------
    [Pre+2007] Press, W H, Teukolsky, S A, Vetterling, W T, and Flannery, B P.
    *Numerical Recipes in C - The Art of Scientific Computing* (Volume 3).
    Cambridge University Press, 2007.
    """
    arr, backend = get_array_and_backend(arr)
    backend : ArrayBackend
    dtype = backend.get_dtype_identifier(array=arr)
    if is_real_floating_dtype(arr.dtype):
        arr = arr.astype(complex_dtype(arr.dtype))
    elif not is_complex_dtype(arr.dtype):
        raise ValueError('array data type {} is not a complex floating point '
                         'data type'.format(dtype_repr(arr.dtype)))

    if out is None:
        out = backend.array_constructor(arr, device=arr.device, copy=True)

    elif out is not arr:
        out[:] = arr

    if axes is None:
        axes = list(range(arr.ndim))
    else:
        try:
            axes = [int(axes)]
        except TypeError:
            axes = list(axes)

    shift_list = normalized_scalar_param_list(shift, length=len(axes),
                                              param_conv=bool)

    if sign == '-':
        imag = -1j
    elif sign == '+':
        imag = 1j
    else:
        raise ValueError("`sign` '{}' not understood".format(sign))

    op, op_in = str(op).lower(), op
    if op not in ('multiply', 'divide'):
        raise ValueError("kernel `op` '{}' not understood".format(op_in))

    # Make a list from interp if that's not the case already
    if is_string(interp):
        interp = [str(interp).lower()] * arr.ndim

    out_dtype = _universal_dtype_identifier(out.dtype)

    onedim_arrs = []
    for ax, shift, intp in zip(axes, shift_list, interp):
        x = real_grid.min_pt[ax]
        xi = recip_grid.coord_vectors[ax]

        # First part: exponential array
        onedim_arr = np.exp(imag * x * xi)

        # Second part: interpolation kernel
        len_dft = recip_grid.shape[ax]
        len_orig = real_grid.shape[ax]
        halfcomplex = (len_dft < len_orig)
        odd = len_orig % 2

        fmin = -0.5 if shift else -0.5 + 1.0 / (2 * len_orig)
        if halfcomplex:
            # maximum lies around 0, possibly half a cell left or right of it
            if shift and odd:
                fmax = - 1.0 / (2 * len_orig)
            elif not shift and not odd:
                fmax = 1.0 / (2 * len_orig)
            else:
                fmax = 0.0

        else:  # not halfcomplex
            # maximum lies close to 0.5, half or full cell left of it
            if shift:
                # -0.5 + (N-1)/N = 0.5 - 1/N
                fmax = 0.5 - 1.0 / len_orig
            else:
                # -0.5 + 1/(2*N) + (N-1)/N = 0.5 - 1/(2*N)
                fmax = 0.5 - 1.0 / (2 * len_orig)

        freqs = np.linspace(fmin, fmax, num=len_dft)
        stride = real_grid.stride[ax]

        interp_kernel = _interp_kernel_ft(freqs, intp)
        interp_kernel *= stride

        if op == 'multiply':
            onedim_arr *= interp_kernel
        else:
            onedim_arr /= interp_kernel


        onedim_arrs.append(onedim_arr.astype(out_dtype, copy=AVOID_UNNECESSARY_COPY))

    fast_1d_tensor_mult(out, onedim_arrs, axes=axes, out=out)
    return out


def reciprocal_space(space, axes=None, halfcomplex=False, shift=True,
                     **kwargs):
    """Return the range of the Fourier transform on ``space``.

    Parameters
    ----------
    space : `DiscretizedSpace`
        Real space whose reciprocal is calculated. It must be
        uniformly discretized.
    axes : sequence of ints, optional
        Dimensions along which the Fourier transform is taken.
        Default: all axes
    halfcomplex : bool, optional
        If ``True``, take only the negative frequency part along the last
        axis for. For ``False``, use the full frequency space.
        This option can only be used if ``space`` is a space of
        real-valued functions.
    shift : bool or sequence of bools, optional
        If ``True``, the reciprocal grid is shifted by half a stride in
        the negative direction. With a boolean sequence, this option
        is applied separately to each axis.
        If a sequence is provided, it must have the same length as
        ``axes`` if supplied. Note that this must be set to ``True``
        in the halved axis in half-complex transforms.
        Default: ``True``
    impl : string, optional
        Implementation back-end for the created space.
        Default: ``'numpy'``
    exponent : float, optional
        Create a space with this exponent. By default, the conjugate
        exponent ``q = p / (p - 1)`` of the exponent of ``space`` is
        used, where ``q = inf`` for ``p = 1`` and vice versa.
    dtype : optional
        Complex data type of the created space. By default, the
        complex counterpart of ``space.dtype`` is used.

    Returns
    -------
    rspace : `DiscretizedSpace`
        Reciprocal of the input ``space``. If ``halfcomplex=True``, the
        upper end of the domain (where the half space ends) is chosen to
        coincide with the grid node.
    """
    if not isinstance(space, DiscretizedSpace):
        raise TypeError('`space` {!r} is not a `DiscretizedSpace` instance'
                        ''.format(space))
    if axes is None:
        axes = tuple(range(space.ndim))
    axes = normalized_axes_tuple(axes, space.ndim)

    if not all(space.is_uniform_byaxis[axis] for axis in axes):
        raise ValueError('`space` is not uniformly discretized in the '
                         '`axes` of the transform')

    if halfcomplex and space.field != RealNumbers():
        raise ValueError('`halfcomplex` option can only be used with real '
                         'spaces')

    exponent = kwargs.pop('exponent', None)
    if exponent is None:
        exponent = conj_exponent(space.exponent)

    dtype = kwargs.pop('dtype', None)
    if dtype is None:
        dtype = complex_dtype(space.dtype_identifier)
    else:
        if not is_complex_dtype(dtype):
            raise ValueError('{} is not a complex data type'
                             ''.format(dtype_repr(dtype)))

    impl = kwargs.pop('impl', 'numpy')
    device = kwargs.pop('device', 'cpu')

    # Calculate range
    recip_grid = reciprocal_grid(space.grid, shift=shift,
                                 halfcomplex=halfcomplex, axes=axes)

    # Need to do this for axes of length 1 that are not transformed
    non_axes = [i for i in range(space.ndim) if i not in axes]
    min_pt = {i: space.min_pt[i] for i in non_axes}
    max_pt = {i: space.max_pt[i] for i in non_axes}

    # Make a partition with nodes on the boundary in the last transform axis
    # if `halfcomplex == True`, otherwise a standard partition.
    if halfcomplex:
        max_pt[axes[-1]] = recip_grid.max_pt[axes[-1]]

    part = uniform_partition_fromgrid(recip_grid, min_pt, max_pt)

    # Use convention of adding a hat to represent fourier transform of variable
    axis_labels = list(space.axis_labels)
    for i in axes:
        # Avoid double math
        label = axis_labels[i].replace('$', '')
        axis_labels[i] = '$\\^{{{}}}$'.format(label)

    recip_spc = uniform_discr_frompartition(part, exponent=exponent,
                                            dtype=dtype, impl=impl,
                                            device=device,
                                            axis_labels=axis_labels)

    return recip_spc


if __name__ == '__main__':
    from doctest import testmod, NORMALIZE_WHITESPACE
    testmod(optionflags=NORMALIZE_WHITESPACE)



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test ODL geometry objects for SPECT."""

from __future__ import division
import numpy as np

import odl
from odl.core.util.testutils import all_equal
from odl.applications.tomo.geometry.spect import ParallelHoleCollimatorGeometry


def test_spect():
    det_nx_pix = 64
    det_ny_pix = 64
    det_nx_mm = 4
    det_radius = 200
    n_proj = 180
    det_param = det_nx_mm * det_nx_pix
    dpart = odl.uniform_partition([-det_param, -det_param],
                                  [det_param, det_param],
                                  [det_nx_pix, det_ny_pix])

    apart = odl.uniform_partition(0, 2 * np.pi, n_proj)
    geom = ParallelHoleCollimatorGeometry(apart, dpart, det_radius)
    assert isinstance(geom.detector, odl.applications.tomo.Flat2dDetector)
    assert all_equal(geom.det_radius, det_radius)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""(Fast) Iterative shrinkage-thresholding algorithm."""

from __future__ import print_function, division, absolute_import
import numpy as np


__all__ = ('proximal_gradient', 'accelerated_proximal_gradient')


def proximal_gradient(x, f, g, gamma, niter, callback=None, **kwargs):
    r"""(Accelerated) proximal gradient algorithm for convex optimization.

    Also known as "Iterative Soft-Thresholding Algorithm" (ISTA).
    See `[Beck2009]`_ for more information.

    This solver solves the convex optimization problem::

        min_{x in X} f(x) + g(x)

    where the proximal operator of ``f`` is known and ``g`` is differentiable.

    Parameters
    ----------
    x : ``f.domain`` element
        Starting point of the iteration, updated in-place.
    f : `Functional`
        The function ``f`` in the problem definition. Needs to have
        ``f.proximal``.
    g : `Functional`
        The function ``g`` in the problem definition. Needs to have
        ``g.gradient``.
    gamma : positive float
        Step size parameter.
    niter : non-negative int, optional
        Number of iterations.
    callback : callable, optional
        Function called with the current iterate after each iteration.

    Other Parameters
    ----------------
    lam : float or callable, optional
        Overrelaxation step size. If callable, it should take an index
        (starting at zero) and return the corresponding step size.
        Default: 1.0

    Notes
    -----
    The problem of interest is

    .. math::
        \min_{x \in X} f(x) + g(x),

    where the formal conditions are that
    :math:`f : X \to \mathbb{R}` is proper, convex and lower-semicontinuous,
    and :math:`g : X \to \mathbb{R}` is differentiable and
    :math:`\nabla g` is :math:`1 / \beta`-Lipschitz continuous.

    Convergence is only guaranteed if the step length :math:`\gamma` satisfies

    .. math::
       0 < \gamma < 2 \beta

    and the parameter :math:`\lambda` (``lam``) satisfies

    .. math::
       \sum_{k=0}^\infty \lambda_k (\delta - \lambda_k) = + \infty

    where :math:`\delta = \min \{1, \beta / \gamma\}`.

    References
    ----------
    .. _[Beck2009]: http://epubs.siam.org/doi/abs/10.1137/080716542
    """
    # Get and validate input
    if x not in f.domain:
        raise TypeError('`x` {!r} is not in the domain of `f` {!r}'
                        ''.format(x, f.domain))
    if x not in g.domain:
        raise TypeError('`x` {!r} is not in the domain of `g` {!r}'
                        ''.format(x, g.domain))

    gamma, gamma_in = float(gamma), gamma
    if gamma <= 0:
        raise ValueError('`gamma` must be positive, got {}'.format(gamma_in))

    if int(niter) != niter:
        raise ValueError('`niter` {} not understood'.format(niter))

    lam_in = kwargs.pop('lam', 1.0)
    lam = lam_in if callable(lam_in) else lambda _: float(lam_in)

    # Get the proximal and gradient
    f_prox = f.proximal(gamma)
    g_grad = g.gradient

    # Create temporary
    tmp = x.space.element()

    for k in range(niter):
        lam_k = lam(k)

        # x - gamma grad_g (x)
        tmp.lincomb(1, x, -gamma, g_grad(x))

        # Update x
        x.lincomb(1 - lam_k, x, lam_k, f_prox(tmp))

        if callback is not None:
            callback(x)


def accelerated_proximal_gradient(x, f, g, gamma, niter, callback=None,
                                  **kwargs):
    r"""Accelerated proximal gradient algorithm for convex optimization.

    The method is known as "Fast Iterative Soft-Thresholding Algorithm"
    (FISTA). See `[Beck2009]`_ for more information.

    Solves the convex optimization problem::

        min_{x in X} f(x) + g(x)

    where the proximal operator of ``f`` is known and ``g`` is differentiable.

    Parameters
    ----------
    x : ``f.domain`` element
        Starting point of the iteration, updated in-place.
    f : `Functional`
        The function ``f`` in the problem definition. Needs to have
        ``f.proximal``.
    g : `Functional`
        The function ``g`` in the problem definition. Needs to have
        ``g.gradient``.
    gamma : positive float
        Step size parameter.
    niter : non-negative int, optional
        Number of iterations.
    callback : callable, optional
        Function called with the current iterate after each iteration.

    Notes
    -----
    The problem of interest is

    .. math::
        \min_{x \in X} f(x) + g(x),

    where the formal conditions are that
    :math:`f : X \to \mathbb{R}` is proper, convex and lower-semicontinuous,
    and :math:`g : X \to \mathbb{R}` is differentiable and
    :math:`\nabla g` is :math:`1 / \beta`-Lipschitz continuous.

    Convergence is only guaranteed if the step length :math:`\gamma` satisfies

    .. math::
       0 < \gamma < 2 \beta.

    References
    ----------
    .. _[Beck2009]: http://epubs.siam.org/doi/abs/10.1137/080716542
    """
    # Get and validate input
    if x not in f.domain:
        raise TypeError('`x` {!r} is not in the domain of `f` {!r}'
                        ''.format(x, f.domain))
    if x not in g.domain:
        raise TypeError('`x` {!r} is not in the domain of `g` {!r}'
                        ''.format(x, g.domain))

    gamma, gamma_in = float(gamma), gamma
    if gamma <= 0:
        raise ValueError('`gamma` must be positive, got {}'.format(gamma_in))

    if int(niter) != niter:
        raise ValueError('`niter` {} not understood'.format(niter))

    # Get the proximal
    f_prox = f.proximal(gamma)
    g_grad = g.gradient

    # Create temporary
    tmp = x.space.element()
    y = x.copy()
    t = 1

    for k in range(niter):
        # Update t
        t, t_old = (1 + np.sqrt(1 + 4 * t ** 2)) / 2, t
        alpha = (t_old - 1) / t

        # x - gamma grad_g (y)
        tmp.lincomb(1, y, -gamma, g_grad(y))

        # Store old x value in y
        y.assign(x)

        # Update x
        f_prox(tmp, out=x)

        # Update y
        y.lincomb(1 + alpha, x, -alpha, y)

        if callback is not None:
            callback(x)


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()
